"""
SANS data format
================

Internal data representation for SANS data.
"""

import sys
import datetime
from copy import copy, deepcopy
import json
from io import BytesIO
from collections import OrderedDict

import numpy as np

from reductus.dataflow.lib.uncertainty import Uncertainty
from reductus.dataflow.lib.exporters import exports_HDF5, exports_text
from reductus.vsansred.vsansdata import RawVSANSData, _toDictItem

IS_PY3 = sys.version_info[0] >= 3

IGNORE_CORNER_PIXELS = True

def _b(s):
    if IS_PY3:
        return s.encode('utf-8')
    else:
        return s

def _s(b):
    if IS_PY3:
        return b.decode('utf-8') if hasattr(b, 'decode') else b
    else:
        return b

class RawSANSData(RawVSANSData):
    suffix = ".sans"

class SansData(object):
    """SansData object used for storing values from a sample file (not div/mask).
       Stores the array of data as a Uncertainty object (detailed in uncertainty.py)
       Stores all metadata
       q, qx, qy, theta all get updated with values throughout the reduction process
       Tsam and Temp are just used for storage across modules (in wireit)
    """
    def __init__(self, data=None, metadata=None, q=None, qx=None, qy=None, aspect_ratio=1.0,
                 xlabel="X", ylabel="Y",
                 theta=None, Tsam=None, Temp=None, attenuation_corrected=False):
        if isinstance(data, np.ndarray):
            # Data is counts, so variance is counts.  Set variance on zero counts to 1
            self.data = Uncertainty(data, data + (data==0))
        else:
            self.data = data
        self.metadata = metadata if metadata is not None else {}
        # There are many places where q was not set, i think i fixed most,
        # but there might be more; be wary
        self.q = q
        self.qx = qx
        self.qy = qy
        self.qx_max = None
        self.qy_max = None
        self.qx_min = None
        self.qy_min = None
        self.xlabel = xlabel
        self.ylabel = ylabel
        self.aspect_ratio = aspect_ratio
        self.theta = theta
        self.attenuation_corrected = attenuation_corrected

        self.Tsam = None #Tsam and Temp are used to store the transmissions for later use
        self.Temp = None
    # Note that I have not defined an inplace subtraction
    def __sub__(self, other):
        result = self.copy()
        if isinstance(other, SansData):
            result.data = self.data - other.data
        else:
            result.data = self.data - other
        return result
    # Actual subtraction
    def __sub1__(self, other):
        result = self.copy()
        if isinstance(other, SansData):
            result.data = self.data - other.data
        else:
            result.data = self.data - other
        return result
    def __add__(self, other):
        result = self.copy()
        if isinstance(other, SansData):
            result.data = self.data + other.data
        else:
            result.data = self.data + other
        return result
    def __rsub__(self, other):
        result = self.copy()
        result.data = other - self.data
        return result
    def __truediv__(self, other):
        result = self.copy()
        if isinstance(other, SansData):
            result.data = self.data/other.data
        else:
            result.data = self.data/other
        return result
    def __mul__(self, other):
        result = self.copy()
        if isinstance(other, SansData):
            result.data = self.data * other.data
        else:
            result.data = self.data * other
        return result

    def copy(self):
        return SansData(copy(self.data), deepcopy(self.metadata),
                        q=copy(self.q), qx=copy(self.qx), qy=copy(self.qy),
                        theta=copy(self.theta), aspect_ratio=self.aspect_ratio,
                        xlabel=self.xlabel, ylabel=self.ylabel,
                        attenuation_corrected=self.attenuation_corrected)

    def __copy__(self):
        return self.copy()

    #def __str__(self):
        #return self.data.x.__str__()
    #def __repr__(self):
        #return self.__str__()

    def get_plottable(self):
        data = self.data.x.astype("float")
        xdim = data.shape[0]
        ydim = data.shape[1]
        if not (np.abs(data) > 1e-10).any():
            zmin = 0.0
            zmax = 1.0
        else:
            zmin = data[data > 1e-10].min()
            if IGNORE_CORNER_PIXELS:
                mask = np.ones(data.shape, dtype='bool')
                mask[0, 0] = mask[-1, 0] = mask[-1, -1] = mask[0, -1] = 0.0
                zmax = data[mask].max()
            else:
                zmax = data.max()
        if self.qx is None or self.qy is None:
            xmin = 0.0
            xmax = 128
            ymin = 0.0
            ymax = 128
        else:
            xmin = self.qx_min if self.qx_min is not None else self.qx.min()
            xmax = self.qx_max if self.qx_max is not None else self.qx.max()
            ymin = self.qy_min if self.qy_min is not None else self.qy.min()
            ymax = self.qy_max if self.qy_max is not None else self.qy.max()
        plottable_data = {
            'entry': self.metadata['entry'],
            'type': '2d',
            'z':  [data.flatten().tolist()],
            'title': _s(self.metadata['run.filename'])+': ' + _s(self.metadata['sample.labl']),
            #'metadata': self.metadata,
            'options': {
                'fixedAspect': {
                    #'fixAspect': True,
                    #'aspectRatio': 1.0
                }
            },
            'dims': {
                'xmax': xmax,
                'xmin': xmin,
                'ymin': ymin,
                'ymax': ymax,
                'xdim': xdim,
                'ydim': ydim,
                'zmin': zmin,
                'zmax': zmax,
            },
            'xlabel': self.xlabel,
            'ylabel': self.ylabel,
            'zlabel': 'Intensity (I)',
        }
        if self.aspect_ratio is not None:
            plottable_data['options']['fixedAspect'] = {
                'fixAspect': True,
                'aspectRatio': self.aspect_ratio
            }
        return plottable_data

    def get_metadata(self):
        metadata = {}
        metadata.update(_toDictItem(self.metadata, convert_bytes=True))
        return metadata

class Sans1dData(object):
    properties = ['x', 'v', 'dx', 'dv', 'xlabel', 'vlabel', 'xunits', 'vunits', 'xscale', 'vscale', 'metadata', 'fit_function']

    def __init__(self, x, v, dx=0, dv=0, xlabel="", vlabel="", xunits="", vunits="", xscale="linear", vscale="linear", metadata=None, fit_function=None):
        self.x = x
        self.v = v
        self.dx = dx
        self.dv = dv
        self.xlabel = xlabel
        self.vlabel = vlabel
        self.xunits = xunits
        self.vunits = vunits
        self.xscale = xscale
        self.vscale = vscale
        self.metadata = metadata if metadata is not None else {}
        self.fit_function = fit_function

    def to_dict(self):
        props = dict([(p, getattr(self, p, None)) for p in self.properties])
        return _toDictItem(props, convert_bytes=True)

    def get_plottable(self):
        label = "%s: %s" % (self.metadata['run.experimentScanID'], self.metadata['sample.labl'])
        xdata = self.x.tolist()
        ydata = self.v.tolist()
        if not hasattr(self.dx, 'tolist'):
            xerr = np.zeros_like(xdata).tolist()
        else:
            xerr = np.sqrt(self.dx).tolist()
        yerr = np.sqrt(self.dv).tolist()
        data = [[x, y, {"yupper": y+dy, "ylower": y-dy, "xupper": x+dx, "xlower": x-dx}] for x,y,dx,dy in zip(xdata, ydata, xerr, yerr)]
        plottable = {
            "type": "1d",
            "options": {
                "axes": {
                    "xaxis": {"label": self.xlabel},
                    "yaxis": {"label": self.vlabel}
                },
                "series": [{"label": label}]
            },
            "data": [data]
        }
        return plottable

    def get_metadata(self):
        return self.to_dict()

    @exports_text(name="column")
    def to_column_text(self):
        with BytesIO() as fid:
            fid.write(_b("# %s\n" % json.dumps(_toDictItem(self.metadata, convert_bytes=True)).strip("{}")))
            columns = {"columns": [self.xlabel, self.vlabel, "uncertainty", "resolution"]}
            units = {"units": [self.xunits, self.vunits, self.vunits, self.xunits]}
            fid.write(_b("# %s\n" % json.dumps(columns).strip("{}")))
            fid.write(_b("# %s\n" % json.dumps(units).strip("{}")))
            np.savetxt(fid, np.vstack([self.x, self.v, self.dv, self.dx]).T, fmt="%.10e")
            fid.seek(0)
            value = fid.read()

        return {
            "name": getattr(self, "name", "default_name"),
            "entry": self.metadata.get("entry", "default_entry"),
            "file_suffix": ".sans1d.dat",
            "value": value.decode('utf-8'),
        }

class SansIQData(object):
    def __init__(self, I=None, dI=None, Q=None, dQ=None, meanQ=None, ShadowFactor=None, label='', metadata=None):
        self.I = I
        self.dI = dI
        self.Q = Q
        self.dQ = dQ
        self.meanQ = meanQ
        self.ShadowFactor = ShadowFactor
        self.label = label
        self.metadata = metadata if metadata is not None else {}
    
    def get_plottable(self):
        columns = OrderedDict([
            ('Q', {'label': "Q", 'units': "1/Ang", 'errorbars': 'dQ'}),
            ('I', {'label': "I(Q)", 'units': "1/cm", 'errorbars': 'dI'}),
            ('meanQ', {'label': 'Mean Q', 'units': "1/Ang"}),
            ('Q^4', {'label': 'Q^4', 'units': '1/Ang**4'}),
            ('I*Q^4', {'label': 'I * Q^4', 'units': '1/cm * 1/Ang**4'}),
            ('I*Q^2', {'label': 'I * Q^2', 'units': '1/cm * 1/Ang**2'}),
        ])
        datas = OrderedDict([
            ("Q", {"values": self.Q.tolist(), "errorbars": self.dQ.tolist()}),
            ("I", {"values": self.I.tolist(), "errorbars": self.dI.tolist()}),
            ("meanQ", {"values": self.meanQ.tolist(), "errorbars": self.dQ.tolist()}),
            ("Q^4", {"values": (self.meanQ**4).tolist()}),
            ("I*Q^4", {"values": (self.I * self.meanQ**4).tolist()}),
            ("I*Q^2", {"values": (self.I * self.meanQ**2).tolist()}),
        ])
        
        name = self.metadata.get("name", "default_name")
        entry_name = self.metadata.get("entry", "default_entry")
        series = [{"label": "%s:%s" % (name, entry_name)}]
        xcol = "Q"
        ycol = "I"
        plottable = {
            "type": "nd",
            "title": "%s:%s" % (name, entry_name),
            "entry": entry_name,
            "columns": columns,
            "options": {
                "series": series,
                "xtransform": "log",
                "ytransform": "log",
                "axes": {
                    "xaxis": {"label": "%s(%s)" % (columns[xcol]["label"], columns[xcol]["units"])},
                    "yaxis": {"label": "%s(%s)" % (columns[ycol]["label"], columns[ycol]["units"])}
                },
                "xcol": xcol,
                "ycol": ycol,
                "errorbar_width": 0
            },
            "datas": datas
        }
        #print(plottable)
        return plottable

    @exports_text(name="column")
    def to_column_text(self):
        # export to 6-column format compatible with SASVIEW
        # The 6 columns are | Q (1/A) | I(Q) (1/cm) | std. dev. I(Q) (1/cm) | sigmaQ | meanQ | ShadowFactor|
        column_names = ["Q", "I", "dI", "dQ", "meanQ", "ShadowFactor"]
        column_values = [getattr(self, cn) for cn in column_names]
        labels = ["Q (1/A)", "I(Q) (1/cm)", "std. dev. I(Q) (1/cm)", "sigmaQ", "meanQ", "ShadowFactor"]

        with BytesIO() as fid:
            fid.write(_b("# %s\n" % json.dumps(_toDictItem(self.metadata, convert_bytes=True)).strip("{}")))
            fid.write(_b("# %s\n" % json.dumps({"columns": labels}).strip("{}")))
            np.savetxt(fid, np.vstack(column_values).T, fmt="%.10e")
            fid.seek(0)
            value = fid.read()

        return {
            "name": _s(self.metadata.get("name", "default_name")),
            "entry": _s(self.metadata.get("entry", "default_entry")),
            "file_suffix": ".sansIQ.dat",
            "value": value.decode('utf-8'),
        }

    @exports_HDF5(name="NXcanSAS")
    def to_NXcanSAS(self):
        import h5py

        fid = BytesIO()
        h5_item = h5py.File(fid, 'w')

        entry_name = self.metadata.get("entry", "entry")
        nxentry = h5_item.create_group(entry_name)
        nxentry.attrs.update({
            "NX_class": "NXentry",
            "canSAS_class": "SASentry",
            "version": "1.0"
        })
        nxentry["definition"] = "NXcanSAS"
        nxentry["run"] = "<see the documentation>"
        nxentry["title"] = self.metadata["sample.description"]
        datagroup = nxentry.create_group("data")
        datagroup.attrs.update({
            "NX_class": "NXdata",
            "canSAS_class": "SASdata",
            "signal": "I",
            "I_axes": "<see the documentation>",
            "Q_indices": 1
        })
        datagroup["I"] = self.I
        datagroup["I"].attrs["units"] = "1/m"
        datagroup["Q"] = self.Q
        datagroup["Q"].attrs["units"] = "1/nm"

        return {
            "name": _s(self.metadata.get("name", "default_name")),
            "entry": _s(self.metadata.get("entry", "default_entry")),
            "file_suffix": ".sansIQ.nx.h5",
            "value": h5_item,
        }

class Parameters(object):
    def __init__(self, params=None):
        self.params = params

    def get_metadata(self):
        return self.params

    def get_plottable(self):
        #return {"entry": "entry", "type": "params", "params": _toDictItem(self.metadata)}
        return {"entry": "entry", "type": "params", "params": _toDictItem(self.params)}



""" categorize VSANS files """
import re
import dateutil.parser

def SortDataAutomatic(datafiles,
                      excluded=None,
                      TransPanel="MR",
                      ManualHe3Entry=False,
                      New_HE3_Files=None,
                      MuValues=None,
                      TeValues=None,
                      ReassignBlockBeam=None,
                      ReassignEmpty=None):

    if excluded is None:
        excluded = []
        
    BlockBeam = {}
    Configs = {}
    Sample_Names = []
    Scatt = {}
    Trans = {}
    Pol_Trans = {}
    HE3_Trans = {}
    FileNumberList = []

    Trans_filenumbers = {
        "UU": -10,
        "DU": -10,
        "DD": -10,
        "UD": -10,
        "SM": -10,
    }
    Trans_times = {
        "UU": -10,
        "DU": -10,
        "DD": -10,
        "UD": -10,
        "SM": -10,
    }
    filenames = '0'
    record_adam4021 = 0
    record_temp = 0
    CellIdentifier = 0
    HE3OUT_filenumber = -10
    
    files_pairs = [(int(f.metadata['run.instFileNum']), f) for f in datafiles]
    # sort by filenumber
    files_pairs.sort(key=lambda x: x[0])
    # exclude the bad ones
    files_pairs = [p for p in files_pairs if not p[0] in excluded]
    filenumbers = [p[0] for p in files_pairs]
    start_number = filenumbers[0] if len(filenumbers) > 0 else 0
    file_lookup = dict(files_pairs)
    
    for filenumber in filenumbers:
        f = file_lookup[filenumber]
        m = f.metadata
        Count_time = m['run.rtime']
        Descrip = m['sample.labl'].decode()

        if Count_time > 59 and not "Align" in Descrip:
            FileNumberList.append(filenumber)
            print('Reading:', filenumber, ' ', Descrip)
            Listed_Config = m['run.configuration'].decode()
            Sample_Name = Descrip.replace(Listed_Config, '')
            Not_Sample = ['T_UU', 'T_DU', 'T_DD', 'T_UD', 'T_SM', 'T_NP', 'HeIN', 'HeOUT', 'S_UU', 'S_DU', 'S_DD', 'S_UD', 'S_NP', 'S_HeU', 'S_HeD', 'S_SMU', 'S_SMD']
            for i in Not_Sample:
                Sample_Name = Sample_Name.replace(i, '')
            Desired_Temp = m['sample_des.temp']
            if Desired_Temp is not None:
                record_temp = 1
                _stringpart = "{:.4f}".format(Desired_Temp)
                regex = r''
                for s in _stringpart[1:]:
                    regex = r'({}' + regex + r')?'
                substr = regex.format(*list(_stringpart))
                Sample_Name = re.compile(r"{dt}\s*K[,]?".format(dt=substr)).sub('', Sample_Name)
                Temp_String = _stringpart
            else:
                Temp_String = 'na'
            Voltage = m['adam.voltage']
            if Voltage is not None:
                record_adam4021 = 1
                _stringpart = "{:.4f}".format(Voltage)
                regex = r''
                for s in _stringpart[1:]:
                    regex = r'({}' + regex + r')?'
                substr = regex.format(*list(_stringpart))
                Sample_Name = re.compile(r"{dv}\s*V[,]?".format(dv=substr)).sub('', Sample_Name)
                Voltage_String = _stringpart
            else:
                Voltage_String = 'na'
            
            Sample_Name = re.compile(r"\s").sub('', Sample_Name)
            Sample_Base = Sample_Name
            Sample_Name = "{:s}_{:s}_V_{:s}_K".format(Sample_Name, Voltage_String, Temp_String)

            Purpose = m['analysis.filepurpose'].decode() #SCATT, TRANS, HE3
            Intent = m['analysis.intent'].decode() #Sample, Empty, Blocked Beam, Open Beam
            if ReassignBlockBeam is not None and filenumber in ReassignBlockBeam:
                Intent = 'Blocked Beam'
            if ReassignEmpty is not None and filenumber in ReassignEmpty:
                Intent = 'Empty'
            Type = m['sample.labl'].decode()
            End_time = dateutil.parser.parse(m['end_time'])
            TimeOfMeasurement = (End_time.timestamp() - Count_time/2)/3600.0 #in hours
            Trans_Detector = f.detectors['detector_' + TransPanel]
            Trans_Counts = Trans_Detector['integrated_count']['value'][0]
            '''
            #ID = str(f['entry/sample/group_id'][0])
            #trans_mask = Trans_masks['MR']
            #trans_data = np.array(f['entry/instrument/detector_{ds}/data'.format(ds=TransPanel)])
            #trans_data = trans_data*trans_mask
            #Trans_Counts = trans_data.sum()
            '''
            MonCounts = m['run.moncnt']
            Trans_Distance =  Trans_Detector['distance']['value'][0]
            Attenuation = m['run.atten']
            Wavelength = m['resolution.lmda']
            Config = get_unique_config_id(m)
            FrontPolDirection = m['polarization.front'].decode()
            if FrontPolDirection is None:
                FrontPolDirection = 'UNPOLARIZED'
            BackPolDirection = m['polarization.back'].decode()
            if BackPolDirection is None:
                BackPolDirection = 'UNPOLARIZED'

            '''Want to populate Config representative filenumbers on scattering filenumber'''
            if "SCATT" in str(Purpose):
                if Config not in Configs or Configs[Config] == 0:
                    Configs[Config] = filenumber
            else:
                if Config not in Configs:
                    Configs[Config] = 0

            if Config not in BlockBeam:
                BlockBeam[Config] = {'Scatt':{'File' : []}, 'Trans':{'File' : [], 'CountsPerSecond' : []}}
            ''' 
            if len(Configs) < 1:
                Configs = {Config : config_filenumber}
            else:
                if Config not in Configs:
                    Configs.append({Config : config_filenumber})
            if Configs[Config] == 0 and config_filenumber != 0:
                Configs[Config] = config_filenumber
            '''
            _intent = Intent.lower()
            _purpose = Purpose.lower()
            _frontpol = FrontPolDirection.lower()
            _backpol = BackPolDirection.lower()
            if "blocked" in _intent:
                if Config not in BlockBeam:
                        BlockBeam[Config] = {'Scatt':{'File' : []}, 'Trans':{'File' : [], 'CountsPerSecond': []}}
                if "trans" in _purpose or "he3" in _purpose:
                    BlockBeam[Config]['Trans']['File'].append(filenumber)
                    BlockBeam[Config]['Trans']['CountsPerSecond'].append(Trans_Counts/Count_time)
                elif "scatt" in _purpose:
                    BlockBeam[Config]['Scatt']['File'].append(filenumber)
            elif "sample" in _intent or "empty" in _intent or "open" in _intent:
                if Sample_Name not in Sample_Names:
                    Sample_Names.append(Sample_Name)
                Intent_short = Intent # copy
                Intent_short = Intent_short.replace(' Cell', '')
                Intent_short = Intent_short.replace(' Beam', '')
                if "scatt" in _purpose:
                    if Sample_Name not in Scatt:
                        Scatt[Sample_Name] = {'Intent': Intent_short, 'Sample_Base': Sample_Base, 'Config(s)': {}}
                    if Config not in Scatt[Sample_Name]['Config(s)']:
                        Scatt[Sample_Name]['Config(s)'][Config] = {'Unpol': [], 'U': [], 'D': [],'UU': [], 'DU': [], 'DD': [], 'UD': [], 'UU_Time': [], 'DU_Time': [], 'DD_Time': [], 'UD_Time' : []}
                        # dict([(n, []) for n in ['Unpol','U','D','UU','DU','DD','UD','UU_Time','DU_Time','DD_Time','UD_Time']])
                    if "unpolarized" in _frontpol and "unpolarized" in _backpol:
                        Scatt[Sample_Name]['Config(s)'][Config]['Unpol'].append(filenumber)
                    if "up" in _frontpol and "unpolarized" in _backpol:
                        Scatt[Sample_Name]['Config(s)'][Config]['U'].append(filenumber)
                    if "down" in _frontpol and "unpolarized" in _backpol:
                        Scatt[Sample_Name]['Config(s)'][Config]['D'].append(filenumber)
                    
                    xs = None
                    if ManualHe3Entry:
                        pol_match = re.search('S_(UU|DU|DD|UD)$', Type)
                        if pol_match is not None:
                            xs = pol_match.group(1)     
                    else:
                        if "up" in _frontpol and "up" in _backpol:
                            xs = "UU"
                        elif "down" in _frontpol and "up" in _backpol:
                            xs = "DU"
                        elif "down" in _frontpol and "down" in _backpol:
                            xs = "DD"
                        elif "up" in _frontpol and "down" in _backpol:
                            xs = "UD"
                    
                    if xs is not None:
                        Scatt[Sample_Name]['Config(s)'][Config][xs].append(filenumber)
                        Scatt[Sample_Name]['Config(s)'][Config][xs + '_Time'].append(TimeOfMeasurement)

                elif "trans" in _purpose:
                    if Sample_Name not in Trans:
                        Trans[Sample_Name] = {'Intent': Intent_short, 'Sample_Base': Sample_Base, 'Config(s)' : {}}
                    if Config not in Trans[Sample_Name]['Config(s)']:
                        Trans[Sample_Name]['Config(s)'][Config] = {'Unpol_Files': [],
                                                                   'U_Files' : [],
                                                                   'D_Files': [],
                                                                   'Unpol_Trans_Cts': [],
                                                                   'U_Trans_Cts' : [],
                                                                   'D_Trans_Cts' : []}
                    if Sample_Name not in Pol_Trans:
                        Pol_Trans[Sample_Name] = {'T_UU': {'File': [], 'Meas_Time': []},
                                                  'T_DU': {'File': [], 'Meas_Time': []},
                                                  'T_DD': {'File': [], 'Meas_Time': []},
                                                  'T_UD': {'File': [], 'Meas_Time': []},
                                                  'T_SM': {'File': [], 'Meas_Time': []},
                                                  'Config' : []}
                    if "unpolarized" in _frontpol and "unpolarized" in _backpol:
                        Trans[Sample_Name]['Config(s)'][Config]['Unpol_Files'].append(filenumber)
                    if "up" in _frontpol and "unpolarized" in _backpol:
                        Trans[Sample_Name]['Config(s)'][Config]['U_Files'].append(filenumber)
                    if "down" in _frontpol and "unpolarized" in _backpol:
                        Trans[Sample_Name]['Config(s)'][Config]['D_Files'].append(filenumber)

                    xs = None
                    if ManualHe3Entry:
                        pol_match = re.search('T_(UU|DU|DD|UD|SM)$', Type)
                        if pol_match is not None:
                            xs = pol_match.group(1)
                    else:
                        if "up" in _frontpol and "up" in _backpol:
                            xs = "UU"
                        elif "down" in _frontpol and "up" in _backpol:
                            xs = "DU"
                        elif "down" in _frontpol and "down" in _backpol:
                            xs = "DD"
                        elif "up" in _frontpol and "down" in _backpol:
                            xs = "UD"
                        elif "up" in _frontpol and "unpolarized" in _backpol:
                            xs = "SM"
                    
                    if xs is not None:
                        Trans_filenumbers[xs] = filenumber
                        Trans_times[xs] = (End_time.timestamp() - Count_time/2)/3600.0

                        if xs == "SM" and (Trans_filenumbers["SM"] - Trans_filenumbers["UU"] == 4):
                            for txs in ["UU", "DU", "DD", "UD"]:
                                Pol_Trans[Sample_Name]['T_' + txs]['File'].append(Trans_filenumbers[txs])
                                Pol_Trans[Sample_Name]['T_' + txs]['Meas_Time'].append(Trans_times[txs])
                            
                            Pol_Trans[Sample_Name]['T_SM']['File'].append(Trans_filenumbers["SM"])
                            Pol_Trans[Sample_Name]['Config'].append(Config)

                elif "he3" in _purpose:
                    if Type.endswith('HeOUT'):
                        if Sample_Name not in Trans:
                            Trans[Sample_Name] = {'Intent': Intent_short, 'Sample_Base': Sample_Base, 'Config(s)' : {}}
                        if Config not in Trans[Sample_Name]['Config(s)']:
                            Trans[Sample_Name]['Config(s)'][Config] = {'Unpol_Files': [],
                                                                       'U_Files' : [],
                                                                       'D_Files': [],
                                                                       'Unpol_Trans_Cts': [],
                                                                       'U_Trans_Cts' : [],
                                                                       'D_Trans_Cts' : []}
                        Trans[Sample_Name]['Config(s)'][Config]['Unpol_Files'].append(filenumber)
                    if ManualHe3Entry:
                        if New_HE3_Files is not None and filenumber in New_HE3_Files:
                            ScaledOpacity = MuValues[CellIdentifier]
                            TE = TeValues[CellIdentifier]
                            CellTimeIdentifier = (End_time.timestamp() - Count_time)/3600.0
                            HE3Insert_Time = (End_time.timestamp() - Count_time)/3600.0
                            CellIdentifier += 1    
                    else:
                        CellTimeIdentifier = m['polarization.backstart']/3600000 #milliseconds to hours
                        CellName = m['polarization.backname'].decode()
                        CellName = CellName + str(CellTimeIdentifier)
                        if CellTimeIdentifier not in HE3_Trans:
                            HE3Insert_Time = CellTimeIdentifier #milliseconds to hours
                            Opacity = m['he3_back.opacity']
                            ScaledOpacity = Opacity*Wavelength
                            TE = m['he3_back.te']
                    if Type.endswith('HeOUT'):
                        HE3OUT_filenumber = filenumber
                        HE3OUT_config = Config
                        HE3OUT_sample = Sample_Name
                        HE3OUT_attenuators = int(Attenuation)
                    elif Type.endswith('HeIN'):
                        HE3IN_filenumber = filenumber
                        HE3IN_config = Config
                        HE3IN_sample = Sample_Name
                        HE3IN_attenuators = int(Attenuation)
                        HE3IN_StartTime = (End_time.timestamp() - Count_time/2)/3600.0
                        if HE3OUT_filenumber > 0:
                            if HE3OUT_config == HE3IN_config and HE3OUT_attenuators == HE3IN_attenuators and HE3OUT_sample == HE3IN_sample: #This implies that you must have a 3He out before 3He in of same config and atten
                                if HE3Insert_Time not in HE3_Trans:
                                    HE3_Trans[CellTimeIdentifier] = {'Te' : TE,
                                                                    'Mu' : ScaledOpacity,
                                                                    'Insert_time' : HE3Insert_Time,
                                                                    'Config': [],
                                                                    'HE3_OUT_file': [],
                                                                    'HE3_IN_file': [],
                                                                    'Elasped_time': [],
                                                                    'Cell_name': []}
                                Elasped_time = HE3IN_StartTime - HE3Insert_Time
                                HE3_Trans[CellTimeIdentifier]['Config'].append(HE3IN_config)
                                HE3_Trans[CellTimeIdentifier]['HE3_OUT_file'].append(HE3OUT_filenumber)
                                HE3_Trans[CellTimeIdentifier]['HE3_IN_file'].append(HE3IN_filenumber)
                                HE3_Trans[CellTimeIdentifier]['Elasped_time'].append(Elasped_time)
                                HE3_Trans[CellTimeIdentifier]['Cell_name'].append(CellName)

    output_dict = {
        "Sample_Names": Sample_Names,
        "Configs": Configs,
        "BlockBeam": BlockBeam,
        "Scatt": Scatt,
        "Trans": Trans,
        "Pol_Trans": Pol_Trans,
        "HE3_Trans": HE3_Trans,
        "start_number": start_number,
        "FileNumberList": FileNumberList
    }            
    #return Sample_Names, Configs, BlockBeam, Scatt, Trans, Pol_Trans, HE3_Trans, start_number, FileNumberList
    return output_dict

def ShareSampleBaseTransmissions(Trans, Scatt):
    for Sample in Scatt:
        for Config in Scatt[Sample]['Config(s)']:
            if Sample not in Trans:
                Intent = Scatt[Sample]['Intent']
                Base = Scatt[Sample]['Sample_Base']
                Trans[Sample] = {'Intent': Intent, 'Sample_Base': Base, 'Config(s)': {}}
            if Config not in Trans[Sample]['Config(s)']:
                Trans[Sample]['Config(s)'][Config] = {'Unpol_Files': [], 'U_Files' : [], 'D_Files': [],'Unpol_Trans_Cts': [], 'U_Trans_Cts' : [], 'D_Trans_Cts' : []}
    UnpolBases = {}
    UnpolAssociatedTrans = {}
    UpBases = {}
    UpAssociatedTrans = {}
    for Sample in Trans:
        Base = Trans[Sample]['Sample_Base']
        if 'Config(s)' in Trans[Sample]:
            for Config in Trans[Sample]['Config(s)']:
                if len(Trans[Sample]['Config(s)'][Config]['Unpol_Files']) > 0:
                    fn = Trans[Sample]['Config(s)'][Config]['Unpol_Files'][0]
                    if Config not in UnpolBases:
                        UnpolBases[Config] = [Base]
                        UnpolAssociatedTrans[Config] = [fn]
                    elif Base not in UnpolBases[Config]:
                        UnpolBases[Config].append(Base)
                        UnpolAssociatedTrans[Config].append(fn)
                if len(Trans[Sample]['Config(s)'][Config]['U_Files']) > 0:
                    fn = Trans[Sample]['Config(s)'][Config]['U_Files'][0]
                    if Config not in UpBases:
                        UpBases[Config] = [Base]
                        UpAssociatedTrans[Config] = [fn]
                    elif Base not in UpBases[Config]:
                        UpBases[Config].append(Base)
                        UpAssociatedTrans[Config].append(fn)
    for Sample in Trans:
        Base = Trans[Sample]['Sample_Base']
        if 'Config(s)' in Trans[Sample]:
            for Config in Trans[Sample]['Config(s)']:
                if not len(Trans[Sample]['Config(s)'][Config]['Unpol_Files']) > 0:
                    if Config in UnpolBases:
                        if Base in UnpolBases[Config]:
                            for i in [i for i,x in enumerate(UnpolBases[Config]) if x == Base]:
                                Trans[Sample]['Config(s)'][Config]['Unpol_Files'] = [UnpolAssociatedTrans[Config][i]]
                if not len(Trans[Sample]['Config(s)'][Config]['U_Files']) > 0:
                    if Config in UpBases:
                        if Base in UpBases[Config]:
                            for i in [i for i,x in enumerate(UpBases[Config]) if x == Base]:
                                Trans[Sample]['Config(s)'][Config]['U_Files'] = [UpAssociatedTrans[Config][i]]
    
    return Trans

def Process_ScattFiles(Scatt):

    for Sample_Name in Scatt:
        if "empty" in (Scatt[Sample_Name]['Intent']).lower():
            for config in Scatt[Sample_Name]['Config(s)'].values():
                UU_present = (len(config['UU']) > 0) # boolean
                DD_present = (len(config['DD']) > 0)
                DU_present = (len(config['DU']) > 0)
                UD_present = (len(config['UD']) > 0)

                # NSF overrides:
                if not DD_present and UU_present:
                    config['DD'] = config['UU']
                    config['DD_Time'] = config['UU_Time']
                elif DD_present and not UU_present:
                    config['UU'] = config['DD']
                    config['UU_Time'] = config['DD_Time']

                # SF overrides:
                if not UD_present and DU_present:
                    config['UD'] = config['DU']
                    config['UD_Time'] = config['DU_Time']
                elif UD_present and not DU_present:
                    config['DU'] = config['UD']
                    config['DU_Time'] = config['DU_Time']
                    
    return Scatt


def get_unique_config_id(metadata):
    
    Desired_FrontCarriage_Distance = int(metadata['f_det_des.dis']) #in cm
    Desired_MiddleCarriage_Distance = int(metadata['m_det_des.dis']) #in cm
    Wavelength = metadata['resolution.lmda']
    GuideHolder = metadata['resolution.guide']
    if b"CONV" in GuideHolder:
        Guides =  "CvB"
    else:
        GuideNum = int(GuideHolder)
        Guides = "{GuideNum:d}".format(GuideNum=GuideNum) + "Gd"
    '''
    GuideHolder = f['entry/DAS_logs/guide/guide'][0]
    if str(GuideHolder).find("CONV") != -1:
        Guides =  int(0)
    else:
        Guides = int(f['entry/DAS_logs/guide/guide'][0])
    '''
    Configuration_ID = "{Guides}{Desired_FrontCarriage_Distance:d}cmF{Desired_MiddleCarriage_Distance:d}cmM{Wavelength:.4f}Ang".format(**locals())
        
    return Configuration_ID



from numpy import pi, arange, arctan2, sqrt, meshgrid, linspace, sin, cos, array, log, ones, histogram2d, logical_and, zeros_like, ones_like, degrees, mod

def ConvertToCylindrical(array_in, x_min, x_max, y_min, y_max, theta_offset = 0.0, min_r = None, oversample_th = 1.0, oversample_r = 1.0):
    x_axis = linspace(x_min, x_max, array_in.shape[1])
    y_axis = linspace(y_min, y_max, array_in.shape[0])

    x_stepsize = float(x_max - x_min) / (array_in.shape[1] - 1)
    y_stepsize = float(y_max - y_min) / (array_in.shape[0] - 1)

    x,y = meshgrid(x_axis, y_axis)
    # meshgrid makes two new arrays with same dimensions as array_in,
    # but filled with x and y coordinates of each point instead of data

    r_in = sqrt(x**2 + y**2)
    theta_in = arctan2(y, x)
    # these are two more arrays with same dimensions as array_in,
    # but filled with the r, theta coordinates of each point of array_in

    dtheta_min = (1.0/r_in[r_in>0]**2 * sqrt(x[r_in>0]**2 * y_stepsize**2 + y[r_in>0]**2 * x_stepsize)).min() * 180.0/pi
    # dtheta is approximately 1/r^2 * sqrt(x^2 dy + y^2 dx)
    th_step = dtheta_min / oversample_th

    dr_min = (1.0/r_in[r_in>0] * ( abs(x[r_in>0] * x_stepsize) + abs(y[r_in>0] * y_stepsize) )).min()
    # dr is (1/r) * (x dx + y dy)
    r_step = dr_min / oversample_r
    if min_r == None:
        min_r = dr_min

    th_out_min = theta_offset
    th_out_max = theta_offset + 360.0 + th_step
    th_out_axis = arange(th_out_min, th_out_max, th_step )

    r_in_min = r_in.min()
    r_in_max = r_in.max()
    r_out_min = max(r_in.min(), min_r)
    r_out_max = r_in_max
    r_out_axis = arange(r_out_min, r_out_max, r_step)

    th_out, r_out = meshgrid(th_out_axis, r_out_axis)
    output_grid = zeros_like(th_out)
    output_norm = zeros_like(th_out)

    forward_r_list = ( r_in - r_in_min ).flatten()
    forward_th_list = (mod(degrees(theta_in) - theta_offset, 360.0) + theta_offset).flatten()

    reverse_x = (r_out * cos(th_out*pi/180.0))
    reverse_y = (r_out * sin(th_out*pi/180.0))

    inshape = r_in.shape
    reverse_norm = (ones_like(th_out)).flatten()
    hist2d, xedges, yedges = histogram2d(reverse_x.flatten(),reverse_y.flatten(), \
        bins = (inshape[1],inshape[0]), range=((x_min, x_max + x_stepsize),(y_min,y_max+y_stepsize)), weights=reverse_norm)
    input_count = hist2d.T # how many bins in the output map to this bin in the input in the reverse mapping
    input_count += 1.0 # how many bins will be mapped in the forward mapping (all of them)

    forward_weights = (array_in/input_count).flatten()
    forward_norm = (1.0 / input_count).flatten()

    outshape = th_out.shape
    #print('outshape: ', outshape)
    hist2d, xedges, yedges = histogram2d(forward_r_list,forward_th_list, \
        bins = (outshape[0],outshape[1]), range=((r_out_min,r_out_max+r_step),(th_out_min,th_out_max+th_step)), weights=forward_weights)
    output_grid += hist2d # counts in every pixel in input added to corresponding output pixel (forward)
    hist2d, xedges, yedges = histogram2d(forward_r_list,forward_th_list, \
        bins = (outshape[0],outshape[1]), range=((r_out_min,r_out_max+r_step),(th_out_min,th_out_max+th_step)), weights=forward_norm)
    output_norm += hist2d # weight of every pixel in input added to corresponding output-weight pixel (forward)

    reverse_x_lookup = ((reverse_x - x_min) / x_stepsize).astype(int)
    reverse_y_lookup = ((reverse_y - y_min) / y_stepsize).astype(int)

    reverse_mask = logical_and((reverse_x_lookup >=0), (reverse_x_lookup < r_in.shape[1]))
    reverse_mask = logical_and(reverse_mask, (reverse_y_lookup >= 0))
    reverse_mask = logical_and(reverse_mask, (reverse_y_lookup < r_in.shape[0]))

    reverse_x_lookup = reverse_x_lookup[reverse_mask]
    reverse_y_lookup = reverse_y_lookup[reverse_mask]

    reverse_r_list = r_out[reverse_mask]
    reverse_th_list = th_out[reverse_mask]

    reverse_weights = (array_in/input_count)[reverse_y_lookup, reverse_x_lookup]
    reverse_norm = (1.0/input_count)[reverse_y_lookup, reverse_x_lookup]

    # counts from corresponding pixel in input added to every output pixel (reverse lookup)
    output_grid[reverse_mask] += reverse_weights

    # weight of corresponding pixel in input added to every output-weight pixel (reverse lookup)
    output_norm[reverse_mask] += reverse_norm

    normalized = zeros_like(output_grid)
    normalized[reverse_mask] = output_grid[reverse_mask] / output_norm[reverse_mask]
    output = output_grid.copy()
    extent = [th_out.min(), th_out.max(), r_out.min(), r_out.max()]

    return output_grid, output_norm, normalized, extent

if __name__ == '__main__':
    from pylab import imshow, show, xlabel, ylabel, colorbar, figure, title
    x, y = meshgrid(linspace(-50., 150., 201), linspace(-20., 30., 51))
    r = sqrt(x**2 + y**2)
    th = arctan2(y, x)
    d = sin(r*pi/25.0)**2
    #d = cos(th)**2 * 1.0/(r + 1.0)**2
    array_out, data_mask, normalized, extent = ConvertToCylindrical(d, -50, 150, -20, 30)
    # transposing the meshgrid because meshgrid(x,y) has shape (len(y), len(x))

    figure()
    imshow(array_out, aspect = 'auto', origin='lower', extent = extent, interpolation="nearest")
    xlabel(r'$\theta ({}^{\circ})$', size='large')
    ylabel(r'$r$', size='large')
    title(r'$\sin^2(r)$ in cylindrical coords', size='large')
    colorbar()

    figure()
    imshow(data_mask, aspect = 'auto', origin='lower', extent=extent, interpolation="nearest")
    xlabel(r'$\theta ({}^{\circ})$', size='large')
    ylabel(r'$r$', size='large')
    title(r'norm for $\sin^2(r)$ in cylindrical coords', size='large')
    colorbar()

    figure()
    imshow(normalized, aspect = 'auto', origin='lower', extent=extent, interpolation="nearest")
    xlabel(r'$\theta ({}^{\circ})$', size='large')
    ylabel(r'$r$', size='large')
    title(r'normalized transform for $\sin^2(r)$ in cylindrical coords', size='large')
    colorbar()

    figure()
    imshow(d, origin='lower', extent=[-50,150,-20,30], interpolation="nearest")
    xlabel(r'$x$', size='large')
    ylabel(r'$y$', size='large')
    title(r'$\sin^2(r)$ in rectangular coords', size='large')
    colorbar()

    a2, m2, n2, e2 = ConvertToCylindrical(d, -50, 150, -20, 30, theta_offset=-90.0)
    figure()
    imshow(a2, aspect = 'auto', origin='lower', extent=e2, interpolation="nearest")
    xlabel(r'$\theta ({}^{\circ})$', size='large')
    ylabel(r'$r$', size='large')
    title(r'$\sin^2(r)$ in cylindrical coords with -90.0 deg offset', size='large')
    colorbar()

    print("show?")
    figure()
    imshow(n2, aspect = 'auto', origin='lower', extent=e2, interpolation="nearest")
    #imshow(m2.T, origin='lower', extent=e2, interpolation="nearest")
    xlabel(r'$\theta ({}^{\circ})$', size='large')
    ylabel(r'$r$', size='large')
    title(r'normalized $\sin^2(r)$ in cylindrical coords with -90.0 deg offset', size='large')
    colorbar()
    show()



"""
SANS data format
================

Internal data representation for SANS data.
"""

import sys
import datetime
from copy import copy, deepcopy
from collections import OrderedDict
import json
from io import BytesIO

import numpy as np

from reductus.dataflow.lib.uncertainty import Uncertainty
from reductus.dataflow.lib.exporters import exports_HDF5, exports_text

IS_PY3 = sys.version_info[0] >= 3

IGNORE_CORNER_PIXELS = True

short_detectors = ["B", "MB", "MT", "ML", "MR", "FT", "FB", "FL", "FR"]

def _b(s):
    if IS_PY3:
        return s.encode('utf-8') if hasattr(s, 'encode') else s
    else:
        return s

def _s(b):
    if IS_PY3:
        return b.decode('utf-8') if hasattr(b, 'decode') else b
    else:
        return b

class RawVSANSData(object):
    suffix = ".vsans"
    def __init__(self, metadata, detectors=None):
        self.metadata = metadata
        self.metadata['name'] = metadata['run.filename']
        self.detectors = detectors

    def todict(self):
        return _toDictItem(self.metadata)

    def get_plottable(self):
        #return {"entry": "entry", "type": "params", "params": _toDictItem(self.metadata)}
        to_exclude = ["fileinfo"]
        metadata = self.metadata.copy()
        for key in to_exclude:
            metadata.pop(key, None)
        return {"entry": "entry", "type": "metadata", "values": _toDictItem(metadata, convert_bytes=True)}

    def get_metadata(self):
        return _toDictItem(self.metadata, convert_bytes=True)

    @exports_text(name="metadata")
    def to_column_text(self):
        output = json.dumps(_toDictItem(self.metadata, convert_bytes=True))
        return {
            "name": _s(self.metadata.get("name", "default_name")),
            "entry": _s(self.metadata.get("entry", "default_entry")),
            "file_suffix": self.suffix + "metadata.json",
            "value": output,
        }

class RawVSANSHe3Data(RawVSANSData):
    pass

def _toDictItem(obj, convert_bytes=False):
    if isinstance(obj, np.integer):
        obj = int(obj)
    elif isinstance(obj, np.floating):
        obj = float(obj)
    elif isinstance(obj, np.ndarray):
        obj = obj.tolist()
    elif isinstance(obj, Uncertainty):
        obj = _toDictItem({'x': obj.x, 'variance': obj.variance})
    elif isinstance(obj, datetime.datetime):
        obj = [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]
    elif isinstance(obj, list):
        obj = [_toDictItem(a, convert_bytes=convert_bytes) for a in obj]
    elif isinstance(obj, dict):
        obj = OrderedDict([(k, _toDictItem(v, convert_bytes=convert_bytes)) for k, v in obj.items()])
    elif isinstance(obj, bytes) and convert_bytes == True:
        obj = obj.decode()
    return obj

class VSansData(object):
    """VSansData object used for storing values from a sample file (not div/mask).
       Stores the array of data as a Uncertainty object (detailed in uncertainty.py)
       Stores all metadata
       q, qx, qy, theta all get updated with values throughout the reduction process
       Tsam and Temp are just used for storage across modules (in wireit)
    """
    properties = ['metadata', 'detectors']

    def __init__(self, metadata=None, detectors=None):
        self.metadata = metadata if metadata is not None else {}
        self.detectors = detectors if detectors is not None else {}
    
    def copy(self):
        return self.__class__(deepcopy(self.metadata), deepcopy(self.detectors))

    def __copy__(self):
        return self.copy()

    #def __str__(self):
        #return self.data.x.__str__()
    #def __repr__(self):
        #return self.__str__()

    def get_plottable(self):
        datasets = []
        zmin = +np.inf
        zmax = -np.inf
        series = []
        series_color = "red"
        for sn in short_detectors:
            detname = 'detector_{short_name}'.format(short_name=sn)
            det = self.detectors.get(detname, None)
            if det is None:
                continue
            series.append({"label": sn, "color": series_color})
            corrected = det['data'].x
            dimX, dimY = corrected.shape
            xaxis = det[self.xaxisname]
            yaxis = det[self.yaxisname]
            xmax = xaxis.max()
            xmin = xaxis.min()
            xstep = (xmax - xmin) / (max(dimX-1, 1))
            ymax = yaxis.max()
            ymin = yaxis.min()
            ystep = (ymax - ymin) / (max(dimY-1, 1))
            zmax = np.max([corrected.max(), zmax])
            zmin = np.min([corrected.min(), zmin])

            new_dataset = {
                "data": corrected.ravel('C'),
                "dims": {
                    "xmin": xmin - xstep/2.0,
                    "xmax": xmax + xstep/2.0,
                    "xdim": dimX,
                    "ymin": ymin - ystep/2.0,
                    "ymax": ymax + ystep/2.0,
                    "ydim": dimY,
                }
            }
            if 'shadow_mask' in det:
                new_dataset['mask'] = det['shadow_mask'].astype(float).ravel('C')
            datasets.append(new_dataset)
        
        data_2d = {
            "dims": {
                "zmin": zmin,
                "zmax": zmax,
            },
            "type": "2d_multi",
            "title": _b(self.metadata["run.filename"]) + b":" + _b(self.metadata["sample.labl"]),
            #"z": [output_grid.T.tolist()],
            "datasets": datasets,
            "ztransform": "log",
            "xlabel": self.xaxisname,
            "ylabel": self.yaxisname,
            "zlabel": "intensity",
            "options": {
                "fixedAspect": {
                    "fixAspect": True,
                    "aspectRatio": 1.0
                },
                "legend": {"show": True},
                "mask": {
                    "show": True,
                    "method": "overlay",
                    "overlay_color": "lightGreen",
                    "overlay_opacity": 0.8
                },
                "series": series
            },
            "metadata": self.metadata.copy()
        }

        return _toDictItem(data_2d, convert_bytes=True)

        #return {"entry": "entry", "type": "metadata", "values": _toDictItem(self.metadata)}

    def get_metadata(self):
        metadata = {}
        metadata.update(_toDictItem(self.metadata, convert_bytes=True))
        return metadata

    def todict(self):
        props = dict([(p, getattr(self, p, None)) for p in self.properties])
        return _toDictItem(props)


class VSansDataRealSpace(VSansData):
    def __init__(self, metadata=None, detectors=None):
        self.xaxisname = 'X'
        self.yaxisname = 'Y'
        VSansData.__init__(self, metadata=metadata, detectors=detectors)

class VSansDataAngleSpace(VSansData):
    def __init__(self, metadata=None, detectors=None):
        self.xaxisname = 'thetaX'
        self.yaxisname = 'thetaY'
        VSansData.__init__(self, metadata=metadata, detectors=detectors)

class VSansDataQSpace(VSansData):
    def __init__(self, metadata=None, detectors=None):
        self.xaxisname = 'Qx'
        self.yaxisname = 'Qy'
        VSansData.__init__(self, metadata=metadata, detectors=detectors)

    @exports_text(name="column")
    def to_column_text(self):
        # export to 6-column format compatible with SASVIEW
        # Data columns are Qx - Qy - I(Qx,Qy) - err(I) - Qz - SigmaQ_parall - SigmaQ_perp - fSubS(beam stop shadow)
        column_names = ["Qx", "Qy", "I", "dI", "Qz", "SigmaQ_para", "SigmaQ_perp", "ShadowFactor", "detector_id"]
        labels = ["Qx (1/A)", "Qy (1/A)", "I(Q) (1/cm)", "std. dev. I(Q) (1/cm)", "Qz (1/A)", "sigmaQ_para", "sigmaQ_perp", "ShadowFactor", "detector_id"]

        with BytesIO() as fid:
            fid.write(_b("# %s\n" % json.dumps(_toDictItem(self.metadata, convert_bytes=True)).strip("{}")))
            fid.write(_b('# detector_id: {"0":"B","1":"MB","2":"MT","3":"ML","4":"MR","5":"FT","6":"FB","7":"FL","8":FR"}\n'))
            fid.write(_b("# %s\n" % json.dumps({"columns": labels}).strip("{}")))
            for d_id, sn in enumerate(short_detectors):
                detname = 'detector_{short_name}'.format(short_name=sn)
                det = self.detectors.get(detname, None)
                if det is None:
                    continue
                column_length = det['Qx'].size
                column_values = [
                    det['Qx'].ravel('C'),
                    det['Qy'].ravel('C'),
                    (det['data']).x.ravel('C'),
                    np.sqrt((det['data']).variance).ravel('C'),
                    det['Qz'].ravel('C'),
                    np.zeros(column_length, dtype='float'), # not yet calculated
                    np.zeros(column_length, dtype='float'), # not yet calculated
                    np.zeros(column_length, dtype='float'), # not yet calculated
                    np.ones(column_length, dtype='float') * d_id
                ]
                np.savetxt(fid, np.vstack(column_values).T, fmt="%.10e")
            fid.seek(0)
            value = fid.read()

        return {
            "name": _s(self.metadata["run.filename"]),
            "entry": _s(self.metadata.get("entry", "default_entry")),
            "file_suffix": ".vsans2d.dat",
            "value": value.decode(),
        }

    @exports_HDF5(name="NXcanSAS")
    def to_NXcanSAS(self):
        import h5py

        # TODO: Avoid returning open handles to caller.
        # Check whether closing h5_item frees fid, or whether it is kept
        # around indefinitely leading to a memory leak.
        fid = BytesIO()
        h5_item = h5py.File(fid, 'w')

        entry_name = self.metadata.get("entry", "entry")
        nxentry = h5_item.create_group(entry_name)
        nxentry.attrs.update({
            "NX_class": "NXentry",
            "canSAS_class": "SASentry",
            "version": "1.0"
        })
        nxentry["definition"] = "NXcanSAS"
        nxentry["run"] = "<see the documentation>"
        nxentry["title"] = self.metadata["sample.description"]

        instrument = nxentry.create_group("instrument")
        total_length = 0
        for d_id, sn in enumerate(short_detectors):
            detname = 'detector_{short_name}'.format(short_name=sn)
            det = self.detectors.get(detname, None)
            if det is None:
                continue
            g = instrument.create_group(detname)
            g.attrs["NX_class"] = "NXdetector"
            g["data"] = det["data"].x
            g["data_variance"] = det["data"].variance
            g["Qx"] = det["Qx"]
            g["Qy"] = det["Qy"]
            g["Qz"] = det["Qz"]
            total_length += det["data"].x.size

        I = np.empty((total_length), dtype=float)
        dI = np.empty((total_length), dtype=float)
        Q = np.empty((3,total_length), dtype=float)

        data_cursor = 0
        for d_id, sn in enumerate(short_detectors):
            detname = 'detector_{short_name}'.format(short_name=sn)
            det = self.detectors.get(detname, None)
            if det is None:
                continue
            I_new = det["data"].x
            dI_new = det["data"].variance
            I[data_cursor:data_cursor + I_new.size] = I_new.ravel("C")
            dI[data_cursor:data_cursor + I_new.size] = dI_new.ravel("C")
            for qi, qn in enumerate(["Qx", "Qy", "Qz"]):
                Q[qi, data_cursor:data_cursor + I_new.size] = det[qn].ravel("C")

        # TODO: convert this to a view of detectors using h5py.VirtualSource?
        datagroup = nxentry.create_group("data")
        datagroup.attrs.update({
            "NX_class": "NXdata",
            "canSAS_class": "SASdata",
            "signal": "I",
            "I_axes": "<see the documentation>",
            "Q_indices": [1,2,3]
        })
        datagroup["I"] = I
        datagroup["I"].attrs["units"] = "1/m"
        datagroup["I"].attrs["uncertainties"] = "Idev"
        datagroup["Idev"] = dI
        datagroup["Idev"].attrs["units"] = datagroup["I"].attrs["units"]
        datagroup["Q"] = Q
        datagroup["Q"].attrs["units"] = "1/nm"

        return {
            "name": _s(self.metadata.get("name", "default_name")),
            "entry": _s(self.metadata.get("entry", "default_entry")),
            "file_suffix": ".sansIQ.nx.h5",
            "value": h5_item,
        }

class VSans1dData(object):
    properties = ['x', 'v', 'dx', 'dv', 'xlabel', 'vlabel', 'xunits', 'vunits', 'xscale', 'vscale', 'metadata', 'fit_function']

    def __init__(self, x, v, dx=0, dv=0, xlabel="", vlabel="", xunits="", vunits="", xscale="linear", vscale="linear", metadata=None, fit_function=None):
        self.x = x
        self.v = v
        self.dx = dx
        self.dv = dv
        self.xlabel = xlabel
        self.vlabel = vlabel
        self.xunits = xunits
        self.vunits = vunits
        self.xscale = xscale
        self.vscale = vscale
        self.metadata = metadata if metadata is not None else {}
        self.fit_function = fit_function

    def todict(self):
        props = dict([(p, getattr(self, p, None)) for p in self.properties])
        return _toDictItem(props)

    def get_plottable(self):
        label = self.metadata.get('title', "unknown"),
        xdata = self.x.tolist()
        ydata = self.v.tolist()
        yerr = self.dv.tolist()
        data = [[x, y, {"yupper": y+dy, "ylower": y-dy, "xupper": x, "xlower": x}] for x,y,dy in zip(xdata, ydata, yerr)]
        plottable = {
            "type": "1d",
            "options": {
                "axes": {
                    "xaxis": {"label": self.xlabel},
                    "yaxis": {"label": self.vlabel}
                },
                "series": [{"label": label}]
            },
            "data": [data]
        }
        return plottable

    def get_metadata(self):
        return _toDictItem(self.metadata)

    @exports_text(name="column")
    def to_column_text(self):
        with BytesIO() as fid:
            fid.write(_b("# %s\n" % json.dumps(_toDictItem(self.metadata, convert_bytes=True)).strip("{}")))
            columns = {"columns": [self.xlabel, self.vlabel, "uncertainty", "resolution"]}
            units = {"units": [self.xunits, self.vunits, self.vunits, self.xunits]}
            fid.write(_b("# %s\n" % json.dumps(columns).strip("{}")))
            fid.write(_b("# %s\n" % json.dumps(units).strip("{}")))
            np.savetxt(fid, np.vstack([self.x, self.v, self.dv, self.dx]).T, fmt="%.10e")
            fid.seek(0)
            value = fid.read()

        return {
            "name": getattr(self, "name", "default_name"),
            "entry": getattr(self.metadata, "entry", "default_entry"),
            "file_suffix": ".vsans1d.dat",
            "value": value.decode('utf-8'),
        }

class Sans1dData(object):
    properties = ['x', 'v', 'dx', 'dv', 'xlabel', 'vlabel', 'xunits', 'vunits', 'xscale', 'vscale', 'metadata', 'fit_function']

    def __init__(self, x, v, dx=0, dv=0, xlabel="", vlabel="", xunits="", vunits="", xscale="linear", vscale="linear", metadata=None, fit_function=None):
        self.x = x
        self.v = v
        self.dx = dx
        self.dv = dv
        self.xlabel = xlabel
        self.vlabel = vlabel
        self.xunits = xunits
        self.vunits = vunits
        self.xscale = xscale
        self.vscale = vscale
        self.metadata = metadata if metadata is not None else {}
        self.fit_function = fit_function

    def todict(self):
        props = dict([(p, getattr(self, p, None)) for p in self.properties])
        return _toDictItem(props)

    def get_plottable(self):
        label = "%s: %s" % (self.metadata['run.instFileNum'], self.metadata['sample.labl'])
        xdata = self.x.tolist()
        ydata = self.v.tolist()
        yerr = self.dv.tolist()
        data = [[x, y, {"yupper": y+dy, "ylower": y-dy, "xupper": x, "xlower": x}] for x,y,dy in zip(xdata, ydata, yerr)]
        plottable = {
            "type": "1d",
            "options": {
                "axes": {
                    "xaxis": {"label": self.xlabel},
                    "yaxis": {"label": self.vlabel}
                },
                "series": [{"label": label}]
            },
            "data": [data]
        }
        return plottable

    def get_metadata(self):
        return self.to_dict()

    @exports_text(name="column")
    def export(self):
        with BytesIO() as fid:
            fid.write(_b("# %s\n" % json.dumps(_toDictItem(self.metadata)).strip("{}")))
            columns = {"columns": [self.xlabel, self.vlabel, "uncertainty", "resolution"]}
            units = {"units": [self.xunits, self.vunits, self.vunits, self.xunits]}
            fid.write(_b("# %s\n" % json.dumps(columns).strip("{}")))
            fid.write(_b("# %s\n" % json.dumps(units).strip("{}")))
            np.savetxt(fid, np.vstack([self.x, self.v, self.dv, self.dx]).T, fmt="%.10e")
            fid.seek(0)
            value = fid.read()

        return {
            "name": getattr(self, "name", "default_name"),
            "entry": getattr(self.metadata, "entry", "default_entry"),
            "file_suffix": ".sans1d.dat",
            "value": value.decode(),
        }

class Parameters(object):
    def __init__(self, params=None):
        self.params = params

    def get_metadata(self):
        return self.params

    def todict(self):
        return _toDictItem(self.params)

    def get_plottable(self):
        #return {"entry": "entry", "type": "params", "params": _toDictItem(self.metadata)}
        return {"entry": "entry", "type": "params", "params": _toDictItem(self.params)}

class Metadata(OrderedDict):
    def get_plottable(self):
        #return {"entry": "entry", "type": "params", "params": _toDictItem(self.metadata)}
        return {"entry": "entry", "type": "metadata", "values": _toDictItem(self)}

    def get_metadata(self):
        return _toDictItem(self)

    @exports_text(name="json")
    def export(self):
        value = json.dumps(_toDictItem(self, convert_bytes=True))
        return {
            "name": getattr(self, "name", "default_name"),
            "entry": getattr(self, "entry", "default_entry"),
            "file_suffix": ".vsans.metadata.json",
            "value": output,
        }



## MetaArray
## Luke Campagnola, 2008
## lcampagn@email.unc.edu
## Class for storing n-dimensional data sets with per-axis meta information
## Free for any kind of use.

import types
from copy import deepcopy

from numpy import ndarray, array, empty, fromstring, concatenate

def axis(name=None, cols=None, values=None, units=None):
  """Convenience function for generating axis descriptions when defining MetaArrays

  Example::

    MetaArray([...], info=[
      axis('Time', values=[0.0, 0.1, 0.2, 0.3], units='s'),
      axis('Signal', cols=[('V0', 'V', 'Voltage0'), ('V1', 'V', 'Voltage1'), ('I0', 'A', 'Current0')])
    ])
  """
  ax = {}
  cNameOrder = ['name', 'units', 'title']
  if name is not None:
    ax['name'] = name
  if values is not None:
    ax['values'] = values
  if units is not None:
    ax['units'] = units
  if cols is not None:
    ax['cols'] = []
    for c in cols:
      if not isinstance(c, (list, tuple)):
        c = [c]
      col = {}
      for i in range(0, len(c)):
        col[cNameOrder[i]] = c[i]
      ax['cols'].append(col)
  return ax

def _expand(cls, data, info):
  """
  Inverse of the MetaArray.__reduce__ method, which reconstructs a MetaArray
  from the state returned by __reduce__.
  """
  #print "_expand", cls, info
  return cls(data=data, info=info)
_expand.__safe_for_unpickling__ = True


class MetaArray(ndarray):
  """N-dimensional array with meta data such as axis titles, units, and column names.

  May be initialized with a file name, a tuple representing the dimensions of the array,
  or any arguments that could be passed on to numpy.array()

  The info argument sets the metadata for the entire array. It is composed of a list
  of axis descriptions where each axis may have a name, title, units, and a list of column
  descriptions. An additional dict at the end of the axis list may specify parameters
  that apply to values in the entire array.

  For example, a 2D array of altitude values for a topographical map might look like::

      info=[
        {'name': 'lat', 'title': 'Lattitude'},
        {'name': 'lon', 'title': 'Longitude'},
        {'title': 'Altitude', 'units': 'm'}
      ]

  In this case, every value in the array represents the altitude in feet at the lat, lon
  position represented by the array index. All of the following return the
  value at lat=10, lon=5::

      array[10, 5]
      array['lon':5, 'lat':10]
      array['lat':10][5]

  Now suppose we want to combine this data with another array of equal dimensions that
  represents the average rainfall for each location. We could easily store these as two
  separate arrays or combine them into a 3D array with this description::

      info=[
        {'name': 'vals', 'cols': [
          {'name': 'altitude', 'units': 'm'},
          {'name': 'rainfall', 'units': 'cm/year'}
        ]},
        {'name': 'lat', 'title': 'Lattitude'},
        {'name': 'lon', 'title': 'Longitude'}
      ]

  We can now access the altitude values with array[0] or array['altitude'], and the
  rainfall values with array[1] or array['rainfall']. All of the following return
  the rainfall value at lat=10, lon=5::

      array[1, 10, 5]
      array['lon':5, 'lat':10, 'val': 'rainfall']
      array['rainfall', 'lon':5, 'lat':10]

  Notice that in the second example, there is no need for an extra (4th) axis description
  since the actual values are described (name and units) in the column info for the first axis.
  """

  def __new__(subtype, data=None, file=None, info=None, dtype=None, copy=False):
    #print "meta new"
    if data is not None:
      if isinstance(data, tuple):
        subarr = empty(data, dtype=dtype)
      else:
        subarr = array(data, dtype=dtype, copy=copy)
      subarr = subarr.view(subtype)

      if info is not None:
        try:
          info = list(info)
        except:
          raise Exception("Info must be a list of axis specifications")
        if len(info) < subarr.ndim + 1:
          info.extend([{}]*(subarr.ndim + 1 - len(info)))
        elif len(info) > subarr.ndim + 1:
          raise Exception("Info parameter must be list of length ndim+1 or less.")
        for i in range(0, len(info)):
          if not isinstance(info[i], dict):
            if info[i] is None:
              info[i] = {}
            else:
              raise Exception("Axis specification must be Dict or None")
          if 'values' in info[i]:
            if isinstance(info[i]['values'], list):
              info[i]['values'] = array(info[i]['values'])
            elif not isinstance(info[i]['values'], ndarray):
              raise Exception("Axis values must be specified as list or ndarray")
        #print "setting info"
        subarr._info = info
      elif hasattr(data, '_info'):
        #print "copying info"
        subarr._info = data._info
      else:
        #print "empty info"
        subarr._info = [{}]*(subarr.ndim + 1)

    elif file is not None:
      fd = open(file, 'r')
      meta = ''
      while True:
        line = fd.readline().strip()
        if line == '':
          break
        meta += line
      meta = eval(meta)

      ## read in axis values
      for ax in meta['info']:
        if 'values_len' in ax:
          ax['values'] = fromstring(fd.read(ax['values_len']), dtype=ax['values_type'])
          del ax['values_len']
          del ax['values_type']

      subarr = fromstring(fd.read(), dtype=meta['type'])
      subarr = subarr.view(subtype)
      subarr.shape = meta['shape']
      subarr._info = meta['info']
    else:
      raise TypeError("Cannot create a blank MetaArray")
    # Finally, we must return the newly created object:
    return subarr

  # Note: __array_finalize__ is useless to us because data.view(MetaArray)
  # will not set the axis info to anything useful.  We might as well require
  # the form MetaArray(data, info=info), in which case there is no good
  # reason to initialize info to the empty axes descriptions.
  #def __array_finalize__(self, obj):
  #  subarr._info = [{}]*(subarr.ndim + 1)


  def __reduce__(self):
    return _expand, (self.__class__, self.view(ndarray), self._info)

  def __copy__(self, order=None):
    return self.__class__(data=self.view(ndarray).copy(order),
                          info=self._info, copy=True)

  def __deepcopy__(self, memo=None):
    return self.__class__(data=self.view(ndarray),
                          info=deepcopy(self._info, memo), copy=True)

  def copy(self, order=None):
    return self.__copy__(order=order)

  def __getitem__(self, ind):
    nInd = self._interpretIndexes(ind)
    a = ndarray.__getitem__(self, nInd)
    if type(a) == type(self):  ## generate new info array
      a._info = []
      for i in range(0, len(nInd)):   ## iterate over all axes
        if isinstance(nInd[i], (slice, list)):  ## If the axis is sliced, keep the info but chop if necessary
          a._info.append(self._axisSlice(i, nInd[i]))
      a._info.append(self._info[-1])   ## Tack on extra data
    return a

  def __setitem__(self, ind, val):
    nInd = self._interpretIndexes(ind)
    return ndarray.__setitem__(self, nInd, val)

  def axisValues(self, axis):
    """Return the list of values for an axis"""
    ax = self._interpretAxis(axis)
    if 'values' in self._info[ax]:
      return self._info[ax]['values']
    else:
      raise Exception('Array axis %s (%d) has no associated values.' % (str(axis), ax))

  def xvals(self, axis):
    """Synonym for axisValues()"""
    return self.axisValues(axis)

  def axisUnits(self, axis):
    """Return the units for axis"""
    ax = self._info[self._interpretAxis(axis)]
    if 'units' in ax:
      return ax['units']

  def columnUnits(self, axis, column):
    """Return the units for column in axis"""
    ax = self._info[self._interpretAxis(axis)]
    if 'cols' in ax:
      for c in ax['cols']:
        if c['name'] == column:
          return c['units']
      raise Exception("Axis %s has no column named %s" % (str(axis), str(column)))
    else:
      raise Exception("Axis %s has no column definitions" % str(axis))

  def rowsort(self, axis, key=0):
    """Return this object with all records sorted along axis using key as the index to the values to compare. Does not yet modify meta info."""
    ## make sure _info is copied locally before modifying it!

    keyList = self[key]
    order = keyList.argsort()
    if isinstance(axis, int):
      ind = [slice(None)] * axis
      ind.append(order)
    elif isinstace(axis, str):
      ind = (slice(axis, order),)
    return self[tuple(ind)]

  def append(self, val, axis):
    """Return this object with val appended along axis. Does not yet combine meta info."""
    ## make sure _info is copied locally before modifying it!

    s = list(self.shape)
    axis = self._interpretAxis(axis)
    s[axis] += 1
    n = MetaArray(tuple(s), info=self._info, dtype=self.dtype)
    ind = [slice(None)] * self.ndim
    ind[axis] = slice(None, -1)
    n[tuple(ind)] = self
    ind[axis] = -1
    n[tuple(ind)] = val
    return n

  def extend(self, val, axis):
    """Return the concatenation along axis of this object and val. Does not yet combine meta info."""
    ## make sure _info is copied locally before modifying it!

    axis = self._interpretAxis(axis)
    return MetaArray(concatenate(self, val, axis), info=self._info)

  def infoCopy(self):
    """Return a deep copy of the axis meta info for this object"""
    return deepcopy(self._info)

  def write(self, fileName):
    """Write this object to a file. The object can be restored by calling MetaArray(file=fileName)"""

    meta = { 'shape': self.shape, 'type': str(self.dtype), 'info': self.infoCopy()}
    axstrs = []
    for ax in meta['info']:
      if 'values' in ax:
        axstrs.append(ax['values'].tostring())
        ax['values_len'] = len(axstrs[-1])
        ax['values_type'] = str(ax['values'].dtype)
        del ax['values']
    fd = open(fileName, 'w')
    fd.write(str(meta) + '\n\n')
    for ax in axstrs:
      fd.write(ax)
    fd.write(self.tostring())
    fd.close()

  def _interpretIndexes(self, ind):
    if not isinstance(ind, tuple):
      ind = (ind,)
    nInd = [slice(None)] * self.ndim
    numOk = True  ## Named indices not started yet; numbered sill ok
    for i in range(0, len(ind)):
      (axis, index, isNamed) = self._interpretIndex(ind[i], i, numOk)
      nInd[axis] = index
      if isNamed:
        numOk = False
    return tuple(nInd)

  def _interpretAxis(self, axis):
    if isinstance(axis, str):
      return self._getAxis(axis)
    else:
      return axis

  def _interpretIndex(self, ind, pos, numOk):
    if isinstance(ind, str):
      if not numOk:
        raise Exception("string and integer indexes may not follow named indexes")
      return (pos, self._getIndex(pos, ind), False)
    elif isinstance(ind, slice):
      if isinstance(ind.start, str) or isinstance(ind.stop, str):  ## Not an actual slice!
        axis = self._interpretAxis(ind.start)
        #if isinstance(ind.start, str):
          #axis = self._getAxis(ind.start)
        #else:
          #axis = ind.start
        if isinstance(ind.stop, str):
          index = self._getIndex(axis, ind.stop)
        else:
          index = ind.stop
        return (axis, index, True)
      else:
        return (pos, ind, False)
    elif isinstance(ind, list):
      indList = [self._interpretIndex(i, pos, numOk)[1] for i in ind]
      return (pos, indList, False)
    else:
      if not numOk:
        raise Exception("string and integer indexes may not follow named indexes")
      return (pos, ind, False)

  def _getAxis(self, name):
    for i in range(0, len(self._info)):
      axis = self._info[i]
      if 'name' in axis and axis['name'] == name:
        return i
    raise Exception("No axis named %s.\n  info=%s" % (name, self._info))

  def _getIndex(self, axis, name):
    ax = self._info[axis]
    if ax is not None and 'cols' in ax:
      for i in range(0, len(ax['cols'])):
        if 'name' in ax['cols'][i] and ax['cols'][i]['name'] == name:
          return i
    raise Exception("Axis %d has no column named %s.\n  info=%s" % (axis, name, self._info))

  def _axisCopy(self, i):
    return deepcopy(self._info[i])

  def _axisSlice(self, i, cols):
    if 'cols' in self._info[i] or 'values' in self._info[i]:
      ax = self._axisCopy(i)
      if isinstance(cols, slice):
        if 'cols' in ax:
          ax['cols'] = ax['cols'][cols]
        if 'values' in ax:
          ax['values'] = ax['values'][cols]
      if isinstance(cols, list):
        if 'cols' in ax:
          ax['cols'] = [ax['cols'][i] for i in cols]
        if 'values' in ax:
          ax['values'] = [ax['values'][i] for i in cols]
    else:
      ax = self._info[i]
    return ax

  def __repr__(self):
    return "%s\n    axis info: %s" % (ndarray.__repr__(self), str(self._info))

  def __str__(self):
    return self.__repr__()




"""
Support for different export types.

See :class:`dataflow.core.DataType` for details.
"""
import io
import json
from functools import wraps

import numpy as np

class NumpyEncoder(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, np.ndarray):
            return obj.tolist()
        # Let the base class default method raise the TypeError
        return json.JSONEncoder.default(self, obj)

def json_dumps(obj, compact=True):
    if compact:
        return json.dumps(obj, separators=(",", ":"), cls=NumpyEncoder)
    else:
        return json.dumps(obj, indent=2, cls=NumpyEncoder)

def _build_filename(export, ext="", index=None):
    index_tag = "_{index}".format(index=index) if index is not None else ""
    name = export.get("name", "unknown"+index_tag)
    entry = export.get("entry", "entry")
    ext = export.get("file_suffix", ext)
    filename = "{name}_{entry}{ext}".format(**locals())
    return filename

def json_writer(datasets, export_method=None, template_data=None, concatenate=False):
    exports = [getattr(d, export_method)() for d in datasets]
    outputs = []
    if concatenate and exports:
        data = {
            "template_data": template_data,
            "outputs": [export["value"] for export in exports],
        }
        compact = exports[0].get("compact", False)
        export_string = json_dumps(data, compact=compact)
        #export_string = export_string.encode('utf-8')
        filename = _build_filename(exports[0], ext=".dat", index=None)
        outputs.append({"filename": filename, "value": export_string})
    else:
        for i, export in enumerate(exports):
            data = {
                "template_data": template_data,
                "outputs": [export["value"]],
                #"index": i,   # TODO: do we want the output index?
            }
            compact = export.get("compact", False)
            export_string = json_dumps(data, compact=compact)
            #export_string = export_string.encode('utf-8')
            filename = _build_filename(export, ext=".dat", index=i)
            outputs.append({"filename": filename, "value": export_string})
    return outputs

def text(datasets, export_method=None, template_data=None, concatenate=False):
    header_string = json_dumps(template_data)
    header_string = "#{h}\n".format(h=header_string[1:-1])
    exports = [getattr(d, export_method)() for d in datasets]
    outputs = []
    if concatenate and exports:
        parts = (export['value'] for export in exports)
        export_string = header_string + "\n\n".join(parts)
        #export_string = export_string.encode('utf-8')
        filename = _build_filename(exports[0], ext=".dat")
        outputs.append({"filename": filename, "value": export_string})
    else:
        for i, export in enumerate(exports):
            export_string = header_string + export["value"]
            #export_string = export_string.encode('utf-8')
            filename = _build_filename(export, ext=".dat", index=i)
            outputs.append({"filename": filename, "value": export_string})
    return outputs


NEXUS_VERSION = "4.2.1"
def _set_nexus_attrs(h5_item, filename):
    from h5py.version import hdf5_version
    from . import iso8601

    h5_item.attrs['NX_class'] = "NXroot"
    h5_item.attrs['file_name'] = filename
    h5_item.attrs['file_time'] = iso8601.now()
    h5_item.attrs['HDF5_Version'] = hdf5_version
    h5_item.attrs['NeXus_version'] = NEXUS_VERSION

def hdf(datasets, export_method=None, template_data=None, concatenate=False):
    import h5py

    header_string = json_dumps(template_data)
    exports = [getattr(d, export_method)() for d in datasets]
    outputs = []
    if concatenate and exports:
        filename = _build_filename(exports[0], ext=".hdf5", index=None)
        fid = io.BytesIO()
        container = h5py.File(fid, 'w')
        _set_nexus_attrs(container, filename)
        container.attrs["template_def"] = header_string
        for export in exports:
            h5_item = export["value"]
            group_to_copy = list(h5_item.values())[0]
            group_name = "%s_%s" % (export["name"], export["entry"])
            for k,v in group_to_copy.items():
                if v.attrs.get("NX_class", "") == "NXprocess":
                    v["template_def"] = header_string
            container.copy(group_to_copy, group_name)
            h5_item.close()
        container.close()
        fid.seek(0)
        outputs.append({"filename": filename, "value": fid.read()})
    else:
        for i, export in enumerate(exports):
            filename = _build_filename(export, ext=".hdf5", index=i)
            h5_item = export["value"]
            _set_nexus_attrs(h5_item, filename)
            h5_item.attrs["template_def"] = header_string
            for entry in h5_item.values():
                if not entry.attrs.get("NX_class", "") == "NXentry":
                    continue
                for k,v in entry.items():
                    if v.attrs.get("NX_class", "") == "NXprocess":
                        v["template_def"] = header_string
            h5_item.flush()
            value = h5_item.id.get_file_image()
            h5_item.close()
            outputs.append({"filename": filename, "value": value})

    return outputs


def png(datasets, export_method=None, template_data=None, concatenate=False):
    # NOTE: concatenate not supported for PNG - keyword ignored.
    from PIL import PngImagePlugin

    header_string = json_dumps(template_data)
    exports = [getattr(d, export_method)() for d in datasets]
    outputs = []
    for i, export in exports:
        image = PngImagePlugin.PngImageFile(export["value"])
        new_metadata = PngImagePlugin.PngInfo()
        existing_metadata = image.text
        for key, value in existing_metadata.items():
            if isinstance(value, PngImagePlugin.iTXt):
                new_metadata.add_itxt(key, value)
            elif isinstance(value, str):
                new_metadata.add_text(key, value)
        new_metadata.add_itxt("reductus_template.json", header_string)
        value = io.BytesIO()
        image.save(value, pnginfo=new_metadata)
        filename = _build_filename(export, ext=".hdf5", index=i)
        outputs.append({"filename": filename, "value": value})

    return outputs


def exports_json(name="json"):
    """
    Decorator for json output file.

    The wrapped function should return a structure containing::

        {
            "name": "base file name",
            "entry": "file entry",
            "file_suffix": ".json",
            "value": {...},  # JSON data (supports numpy arrays as values)
            "compact": True,  # Default is false
        }

    The resulting file will contain::

        {
            "template_data": {...},  # template information
            "outputs": [{...}],  #  list of all outputs in bundle
        }
    """
    def inner_function(f):
        f.exporter = json_writer
        f.export_name = name
        return f
    return inner_function


def exports_text(name="column"):
    def inner_function(f):
        f.exporter = text
        f.export_name = name
        return f
    return inner_function


def exports_HDF5(name="NeXus"):
    """
    Method should return a data block whose "value" key is associated with an
    open h5py File object. The object will be closed after the data is
    extracted for export.
    """
    def inner_function(f):
        f.exporter = hdf
        f.export_name = name
        return f
    return inner_function


def exports_PNG(name="PNG"):
    def inner_function(f):
        f.exporter = png
        f.export_name = name
        return f
    return inner_function



# -*- coding: utf-8 -*-
import re
import json
import os
from time import time, strftime

#import matplotlib.pyplot as plt
import numpy as np

class dcsData(object):
    badbrd = 0
    baddet = 0
    ch_bid = 0
    ch_delay = 0
    ch_dis = 0
    ch_input = 0
    ch_ms = 0
    ch_phase = 0
    ch_res = 0
    ch_slots = 0
    ch_srdenom = 0
    ch_srmode = 0
    ch_wl = 0
    coll_amp = 0
    coll_mean = 0
    coll_osc = 0
    command = 0
    comments = 0
    datamax = 0
    det_dis = 0
    detsum = 0
    duration = 0
    duration_csum = 0
    fc_dis = 0
    grandsum = 0
    he3_csum = 0
    high_csum = 0
    highmax = 0
    highsource = 0
    histodata = 0
    histohigh = 0
    motor_pos = 0
    ncycles = 0
    nframes = 0
    repeats = 0
    resets = 0
    runinfo = 0
    sample_desc = 0
    shutter_stat = 0
    start_date = 0
    startchoice = 0
    stop_date = 0
    tchanlook = 0
    temp_control = 0
    temp_sample = 0
    temp_setpoint = 0
    timsum = 0
    totals = 0
    tsdmin = 0
    user = 0

    # The class "constructor" - It's actually an initializer
    def __init__(self, badbrd, baddet, ch_bid, ch_delay, ch_dis, ch_input, ch_ms, ch_phase, ch_res, ch_slots, ch_srdenom, ch_srmode, ch_wl, coll_amp, coll_mean, coll_osc, command, comments, datamax, det_dis, detsum, duration, duration_csum, fc_dis, grandsum, he3_csum, high_csum, highmax, highsource, histodata, histohigh, motor_pos, ncycles, nframes, repeats, resets, runinfo, sample_desc, shutter_stat, start_date, startchoice, stop_date, tchanlook, temp_control, temp_sample, temp_setpoint, timsum, totals, tsdmin, user ):
        self.badbrd = badbrd
        self.baddet = baddet
        self.ch_bid = ch_bid
        self.ch_delay = ch_delay
        self.ch_dis = ch_dis
        self.ch_input = ch_input
        self.ch_ms = ch_ms
        self.ch_phase = ch_phase
        self.ch_res = ch_res
        self.ch_slots = ch_slots
        self.ch_srdenom = ch_srdenom
        self.ch_srmode = ch_srmode
        self.ch_wl = ch_wl
        self.coll_amp = coll_amp
        self.coll_mean = coll_mean
        self.coll_osc = coll_osc
        self.command = command
        self.comments = comments
        self.datamax = datamax
        self.det_dis = det_dis
        self.detsum = detsum
        self.duration = duration
        self.duration_csum = duration_csum
        self.fc_dis = fc_dis
        self.grandsum = grandsum
        self.he3_csum = he3_csum
        self.high_csum = high_csum
        self.highmax = highmax
        self.highsource = highsource
        self.histodata = histodata
        self.histohigh = histohigh
        self.motor_pos = motor_pos
        self.ncycles = ncycles
        self.nframes = nframes
        self.repeats = repeats
        self.resets = resets
        self.runinfo = runinfo
        self.sample_desc = sample_desc
        self.shutter_stat = shutter_stat
        self.start_date = start_date
        self.startchoice = startchoice
        self.stop_date = stop_date
        self.tchanlook = tchanlook
        self.temp_control = temp_control
        self.temp_sample = temp_sample
        self.temp_setpoint = temp_setpoint
        self.timsum = timsum
        self.totals = totals
        self.tsdmin = tsdmin
        self.user = user

def Elam(lam):
    """
    convert wavelength in angstroms to energy in meV
    """
    return 81.81/lam**2

def Ek(k):
    """
    convert wave-vector in inver angstroms to energy in meV
    """
    return 2.072*k**2

def kE(E):
    return np.sqrt(E/2.072)

def Qfunc(ki, kf, theta):
    """
    evaluate the magnitude of Q from ki, kf, and theta
    theta is the angle between kf and ki, sometimes called 2 theta, units of degrees
    """
    return np.sqrt(  ki**2 + kf**2 - 2*ki*kf*np.cos(theta*np.pi/180)  )

def Ef_from_timechannel(timeChannel, t_SD_min, speedRatDenom, masterSpeed):
    """
    using the parameters
        t_SD_min = minimum sample to detector time
        speedRatDenom = to set FOL chopper speed
        masterSpeed = chopper speed (except for FOL chopper)
    using the variabl
        timeChannel, where I am numbering from 1 <be careful of this convention>
    """
    return 8.41e7 / (t_SD_min + (timeChannel+1)*    (6e4 *(speedRatDenom/masterSpeed))   )**2

def process_raw_dcs(data_path):
    # t0 = time()
    #import matplotlib.pyplot as plt
    os.chdir(data_path) # change working directory
    detInfo = np.genfromtxt('dcs_detector_info.txt', skip_header=1)#, skip_footer=17)
    detToTwoTheta = detInfo[:,9] # 10th column

    #use octave to read the binary-octave dcs files
    os.system('gzip -dc livedata.dcs.gz > livedata.dcs')
    os.system('/usr/bin/octave --eval "load livedata.dcs; save temp_setpoint.txt temp_setpoint; save temp_sample.txt temp_sample; save temp_control.txt temp_control; save shutter_stat.txt shutter_stat; save nframes.txt nframes; save ncycles.txt ncycles ; save motor_pos.txt motor_pos; save duration.txt duration ; save ch_srmode.txt ch_srmode; save comments.txt comments; save coll_osc.txt coll_osc; save histodata.txt histodata;save user.txt user ; save ch_res.txt ch_res; save ch_wl.txt ch_wl; save ch_ms.txt ch_ms; save ch_phase.txt ch_phase; save ch_srdenom.txt ch_srdenom; save tsdmin.txt tsdmin; save stop_date.txt stop_date; save start_date.txt start_date;"')

    #read in the status for the web page
    file = open("status_for_webpage.txt","r")
    status_for_webpage = file.read()
    dry_run_ETA_re = r"TOTAL ESTIMATED DURATION OF SEQUENCE =(.*)"
    dry_run_ETA_re_s = re.search(dry_run_ETA_re, status_for_webpage)
    if dry_run_ETA_re_s:
       dry_run_ETA = dry_run_ETA_re_s.group(1)
    else:
       dry_run_ETA = -99
    dry_run_ETA_wall_re = r"ESTIMATED FINISH TIME IF STARTED IMMEDIATELY   IS (.*)"
    dry_run_ETA_wall_re_s = re.search(dry_run_ETA_wall_re, status_for_webpage)
    if dry_run_ETA_wall_re_s:
       dry_run_ETA_wall = dry_run_ETA_wall_re_s.group(1)
    else:
       dry_run_ETA_wall = -99
    status_for_webpage = status_for_webpage.replace('\n', '<br />')
    file.close()

    latest_file = np.genfromtxt('latest_file.txt', dtype=str)
    thisDcsData = dcsData
    thisDcsData.histodata = np.genfromtxt('histodata.txt')
    thisDcsData.temp_setpoint = np.genfromtxt('temp_setpoint.txt')
    thisDcsData.temp_sample = np.genfromtxt('temp_sample.txt')
    thisDcsData.temp_control = np.genfromtxt('temp_control.txt')
    thisDcsData.motor_pos = np.genfromtxt('motor_pos.txt')
    thisDcsData.ncycles = np.genfromtxt('ncycles.txt')
    thisDcsData.nframes = np.genfromtxt('nframes.txt')
    thisDcsData.shutter_stat = np.genfromtxt('shutter_stat.txt')
    thisDcsData.duration = np.genfromtxt('duration.txt')
    thisDcsData.ch_wl = np.genfromtxt('ch_wl.txt')
    thisDcsData.ch_res = np.genfromtxt('ch_res.txt')
    thisDcsData.coll_osc = np.genfromtxt('coll_osc.txt')
    thisDcsData.ch_ms = np.genfromtxt('ch_ms.txt')
    thisDcsData.ch_phase = np.genfromtxt('ch_phase.txt')
    thisDcsData.ch_srdenom = np.genfromtxt('ch_srdenom.txt')
    thisDcsData.ch_srmode = np.genfromtxt('ch_srmode.txt')
    thisDcsData.tsdmin = np.genfromtxt('tsdmin.txt')
    thisDcsData.stop_date = np.genfromtxt('stop_date.txt', dtype=str)
    thisDcsData.start_date = np.genfromtxt('start_date.txt', dtype=str)
    thisDcsData.user = str(np.genfromtxt('user.txt', dtype=str))
    thisDcsData.comments = str(np.genfromtxt('comments.txt', dtype=str))

    data = np.transpose(thisDcsData.histodata)

    ch_wl = thisDcsData.ch_wl
    Ei = Elam(ch_wl)
    ki = kE(Ei)
    dE = abs(0.5*(-0.10395+0.05616 *Ei+0.00108 *Ei**2)) #take the putative resolution and halve it
    masterSpeed = thisDcsData.ch_ms
    speedRatDenom = thisDcsData.ch_srdenom
    t_SD_min = thisDcsData.tsdmin

    #binning resolution
    Q_max = Qfunc(ki,ki,150)
    Q_min = 0
    E_bins = np.linspace(-Ei, Ei, int(2*Ei/dE)*0.5 )
    Q_bins = np.linspace(Q_min,Q_max,301*0.5)

    #for every point in {timechannel, detectorchannel} space, map into a bin of {E,Q} space
    #remember, data is organized as data[detectorchannel][timechannel]
    i,j = np.indices(data.shape)
    ef = Ef_from_timechannel(j, t_SD_min, speedRatDenom, masterSpeed)
    #print np.shape(data)
    #print np.shape(ki), np.shape(kE(ef)), np.shape(detToTwoTheta[:, None])
    #print detToTwoTheta

    Q_ = Qfunc(ki, kE(ef), detToTwoTheta[:, None])

    E_transfer = Ei-ef
    E_mask = (E_transfer > -Ei)

    EQ_data, xedges, yedges = np.histogram2d(Q_[E_mask], E_transfer[E_mask], bins=(Q_bins, E_bins), range=([Q_min,Q_max], [-Ei, Ei]), weights=data[E_mask])

    #EQ_data = np.transpose(EQ_data)

    #plt.imshow(EQ_data)
    #plt.show()

    stop_date = thisDcsData.stop_date
    start_date = thisDcsData.start_date

    output = {
        "title": "DCS snapshot",
        "dims": {
            "ymin": -Ei,
            "ymax": Ei,
            "ydim": EQ_data.shape[1],
            "xmin": 0,
            "xmax": Q_max,
            "xdim": EQ_data.shape[0],
            "zmin": EQ_data.min(),
            "zmax": EQ_data.max()
        },
        "type": "2d",
        "xlabel": u"|Q| [A⁻¹]",
        "ylabel": "Ei-Ef [meV]",
        "z": [EQ_data.T.tolist()],
        "metadata": {
            "ch_ms": str(thisDcsData.ch_ms),
            "coll_osc": str(thisDcsData.coll_osc),
            "comments": str(thisDcsData.comments),
            #"ch_phase": str(thisDcsData.ch_phase),
            "ch_srdenom": str(thisDcsData.ch_srdenom),
            "ch_srmode": str(thisDcsData.ch_srmode),
            "ch_wl": str(thisDcsData.ch_wl),
            "ch_res": str(thisDcsData.ch_res),
            "duration": str(thisDcsData.duration),
            "latest_file": str(latest_file),
            "motor_pos": str(thisDcsData.motor_pos),
            "ncycles": str(thisDcsData.ncycles),
            "nframes": str(thisDcsData.nframes),
            "shutter_stat": str(thisDcsData.shutter_stat),
            "temp_setpoint": str(thisDcsData.temp_setpoint),
            "temp_sample": str(thisDcsData.temp_sample),
            "temp_control": str(thisDcsData.temp_control),
            "stop_date": str(stop_date[0]+' '+stop_date[1]),
            "user": thisDcsData.user,
            "start_date": str(start_date[0]+' '+start_date[1]),
            #"sample_field": "sample field",
            "tsdmin": str(thisDcsData.tsdmin),
            "dry_run_ETA": dry_run_ETA,
            "dry_run_ETA_wall": dry_run_ETA_wall,
            #"x_dcs_status": r'<strong><th colspan="3">'+status_for_webpage+r'</th></strong>',
            "webdata_updated_last": strftime("%c"),
        },
        "options": {},
        "status":  r'<strong>'+status_for_webpage+r'</strong>',
    }

    #print time()-t0
    return json.dumps([output])


def main():
    open(r'/home/NIST/ncnr/livedata/live_data.json.new', 'w').write(process_raw_dcs(r"/home/NIST/ncnr/livedata/"))
    os.rename('live_data.json.new', 'live_data.json')

if __name__ == "__main__":
    main()



import numpy as np
from .resolution import divergence, divergence_simple

def apply_theta_offset(data, offset):
    data.sample.angle_x += offset
    data.detector.angle_x -= offset


def apply_absolute_angle(data):
    index = (data.sample.angle_x < 0)  # type: np.ndarray
    if np.any(index):
        data.sample.angle_x[index] *= -1.0
        data.detector.angle_x[index] *= -1.0


def apply_back_reflection(data):
    data.sample.angle_x *= -1.0
    data.detector.angle_x *= -1.0


def apply_divergence_simple(data, sample_width):
    # TODO: decide whether we should use Ti or sample.angle internally
    # The difference is that Ti has been shaped to match counts data
    # for a bank of detectors, but sample.angle_x is just the list of
    # sample angles in the scan.

    distance = abs(data.slit1.distance), abs(data.slit2.distance)
    # for slit scans, the sample size and angle are irrelevant and
    # need to be excluded from the calculation of divergence:
    use_sample = data.intent != "intensity"
    if sample_width is None:
        sample_width = data.sample.width

    slits = data.slit1.x, data.slit2.x
    #theta = data.Ti
    theta = data.sample.angle_x
    dtheta = divergence_simple(slits=slits, distance=distance, T=theta,
                               sample_width=sample_width, use_sample=use_sample)
    #print("divergence", theta.shape, dtheta.shape)
    data.angular_resolution = dtheta

def apply_sample_broadening(data, sample_broadening):
    r"""
    Modify angular divergence by a sample broadening (or focussing) amount.

    *sample_broadening* : float | |deg| 1-\ $\sigma$
            Additional divergence caused by sample. This can instead be
            included as a fitting parameter since it simply adds to the
            final $\Delta\theta$

    This makes no attempt to account for beam shadowing by a warped sample at
    really low angles, which will mostly show up as an effective attenuation.
    """
    data.angular_resolution = data.angular_resolution + sample_broadening

def apply_divergence_front_back(data, sample_width):
    """
    Use all four slits when computing divergence.
    """
    # TODO: separate sample_broadening from angular_resolution
    # TODO: decide whether we should use Ti or sample.angle internally
    # The difference is that Ti has been shaped to match counts data
    # for a bank of detectors, but sample.angle_x is just the list of
    # sample angles in the scan.

    slits = data.slit1, data.slit2, data.slit3, data.slit4
    distance = [s.distance for s in slits]
    width = [s.x for s in slits]

    # for slit scans, the sample size and angle are irrelevant and
    # need to be excluded from the calculation of divergence:
    use_sample = data.intent != "intensity"
    if sample_width is None:
        sample_width = data.sample.width

    # Assume Ti = Tf = Td/2
    theta = data.Td/2
    dtheta = divergence(slits=width, distance=distance, T=theta,
                        sample_width=sample_width, use_sample=use_sample)
    #print("divergence", theta.shape, dtheta.shape, dtheta.max())
    data.angular_resolution = dtheta



import os

import numpy as np

from reductus.dataflow.lib import iso8601
from .refldata import ReflData
from .nexusref import data_as, str_data
from .nexusref import NCNRNeXusRefl, load_nexus_entries

def load_entries(filename, file_obj=None, entries=None):
    #print("loading", filename, file_obj)
    return load_nexus_entries(filename, file_obj=file_obj, entries=entries,
                              meta_only=False, entry_loader=MagikHorizontal)

class MagikHorizontal(NCNRNeXusRefl):
    """
    MAGIK horizontal-mode data entry.

    See :class:`refldata.ReflData` for details.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.geometry = "horizontal"
        self.align_intensity = "sample.angle_x"

    def load(self, entry):
        super().load(entry)
        das = entry['DAS_logs']
        n = self.points
        self.slit2.distance = data_as(das, 'horizontalGeom/dS2', '') #das['trajectoryData/_S2'].value
        self.slit1.distance = data_as(das, 'horizontalGeom/d21', '')  + self.slit2.distance
        if 'horizontalGeomBackSlit/dS3' in das:
            self.slit3.distance = data_as(das, 'horizontalGeomBackSlit/dS3', '')

        for k, s in enumerate([self.slit1, self.slit2, self.slit3, self.slit4]):
            s.y = s.x
            s.y_target = s.x_target
            s.x = np.ones_like(s.y) * np.inf
            s.x_target = np.ones_like(s.y) * np.inf

        for k, s in enumerate([self.slit1, self.slit2, self.slit3]):
            x = 'CVertSlit%d/opening'%(k+1)
            if x in das:
                s.x = data_as(das, x, '', rep=n)
                s.x_target = s.x

        tilt = data_as(das, 'horizontalGeom/angleZero', '', rep=n)
        angle = data_as(das, 'horizontalGeom/angle', '', rep=n)
        self.sample.angle_x = angle
        if (self.intent.startswith('rock')):
            self.sample.angle_x -= tilt
        self.sample.angle_x_target = self.sample.angle_x
        if 'horizontalGeomBackSlit/angleMultiplier' in das:
            # then we have a back slit that defines scattering angle
            multiplier = data_as(das, 'horizontalGeomBackSlit/angleMultiplier', '', rep=n)
        else:
            multiplier = 1
        self.detector.angle_x = angle + (angle * multiplier)
        self.detector.angle_x_target = self.detector.angle_x



"""
USANS data loader
================

Load USANS NeXus file into :mod:`usansred.usansdata` data structure.
"""
import io
from collections import OrderedDict
import h5py

from reductus.dataflow.lib import hzf_readonly_stripped as hzf
from reductus.dataflow.lib import unit
from reductus.dataflow.lib.h5_open import h5_open_zip
from reductus.dataflow.lib.strings import _s, _b

from .usansdata import RawData, USansData

metadata_lookup = OrderedDict([
    ("run.filename", "DAS_logs/trajectoryData/fileName"),
    ("analysis.intent", "DAS_logs/trajectoryData/intent"),
    ("sample.name", "DAS_logs/sample/name"),
    ("run.filePrefix", "DAS_logs/trajectoryData/filePrefix"),
    ("run.instFileNum", "DAS_logs/trajectoryData/instFileNum"),
    ("start_time","start_time"),
    ("end_time","end_time"),
    ("eventfile", "DAS_logs/areaDetector/eventFileName"),
])

def readUSANSNexus(input_file, file_obj=None, metadata_lookup=metadata_lookup, det_deadtime=0.0, trans_deadtime=0.0):
    """
    Load all entries from the NeXus file into sans data sets.
    """
    datasets = []
    file = h5_open_zip(input_file, file_obj)
    for entryname, entry in file.items():
        metadata = OrderedDict([
            ("run.filename", _s(entry["DAS_logs/trajectoryData/fileName"][0])),
            ("analysis.intent", _s(entry["DAS_logs/trajectoryData/intent"][0])),
            ("sample.name", _s(entry["DAS_logs/sample/name"][0])),
            ("run.filePrefix", _s(entry["DAS_logs/trajectoryData/filePrefix"][0])),
            ("run.instFileNum", int(entry["DAS_logs/trajectoryData/instFileNum"][0])),
            ("start_time", _s(entry["start_time"][0])),
            ("end_time",_s(entry["end_time"][0])),
            ("entry", _s(entryname)),
            ("dQv", 0.117), # constant of the instrument.  Should it be in the nexus def?
        ])

        counts = entry['DAS_logs/linearDetector/counts'][()]
        countTime = entry['DAS_logs/counter/liveTime'][()]
        trans_counts = entry['DAS_logs/transDetector/counts'][()]
        detCts = (counts / (1.0 - (counts*det_deadtime/countTime[:,None]))).sum(axis=1)
        transCts = (trans_counts / (1.0 - (trans_counts*trans_deadtime/countTime[:,None]))).sum(axis=1)
        monCts = entry['DAS_logs/counter/liveMonitor'][()]
        Q = entry['DAS_logs/analyzerRotation/softPosition'][()]

        dataset = USansData(metadata=metadata, countTime=countTime, detCts=detCts, transCts=transCts, monCts=monCts, Q=Q) 
        datasets.append(dataset)   
    return datasets



#!/usr/bin/env python

import struct
import sys
import numpy
import math
import os

IS_PY3 = (sys.version_info > (3,0))

if IS_PY3:
    def R4toFloat(vaxasstring):
        """Takes a 4 character string represention of a VAX REAL*4 from struct.unpack
                and returns a python float"""	

        if vaxasstring[1] == 0:
            ieeeasstring = bytes([vaxasstring[2],vaxasstring[3],vaxasstring[0],vaxasstring[1]])
            return struct.unpack('<f',ieeeasstring)[0]
        else:
            ieeeasstring = bytes([vaxasstring[2],vaxasstring[3],vaxasstring[0],vaxasstring[1]-1])
            return struct.unpack('<f',ieeeasstring)[0]
else:
    def R4toFloat(vaxasstring):
        """Takes a 4 character string represention of a VAX REAL*4 from struct.unpack
            and returns a python float"""	

        if ord(vaxasstring[1]) == 0:
            ieeeasstring = vaxasstring[2]+vaxasstring[3]+vaxasstring[0]+vaxasstring[1]
            return struct.unpack('<f',ieeeasstring)[0]
        else:
            ieeeasstring = vaxasstring[2]+vaxasstring[3]+vaxasstring[0]+chr(ord(vaxasstring[1])-1)
            return struct.unpack('<f',ieeeasstring)[0]

def readMask(inputfile):
    if hasattr(inputfile, 'read'):
        data = inputfile.read()
    else:
        data = open(inputfile, 'rb').read()

    dataformatstring = '<16384B'
    output = numpy.array(struct.unpack(dataformatstring, data[4:16388]), dtype="float")
    output = numpy.flipud(output.reshape(128,128))
    return output

def readNCNRSensitivity(inputfile):
    
    if hasattr(inputfile, 'read'):
        data = inputfile.read()
    else:
        data = open(inputfile, 'rb').read()
    
    #skip the fake header and just read the data
    #data is 32bit VAX floats
    dataformatstring = '<65600s'
    (rawdatastring,) = struct.unpack(dataformatstring,data[516:])
    
    detdata = numpy.empty(16384)
    
    a = 0
    offset = 0
    for ii in range(16):
        for jj in range(511):
            detdata[a] = R4toFloat(rawdatastring[offset:offset+4])
            a += 1
            offset += 4
        
        offset += 2
        
        for kk in range(510):
	        detdata[a] = R4toFloat(rawdatastring[offset:offset+4])
	        a += 1
	        offset += 4
        
        offset += 2
    
    for ii in range(48):
        detdata[a] = R4toFloat(rawdatastring[offset:offset+4])
        a += 1
        offset += 4
    
    detdata.resize(128,128)
    
    return detdata.T


def readNCNRData(inputfile, file_obj=None):
    
    metadata = {}
    if file_obj is not None and hasattr(file_obj, 'read'):
        data = file_obj.read()
    else:
        data = open(inputfile, 'rb').read()
    
    metadata['run.filename'] = os.path.basename(inputfile)
    
    #filename
    dat = struct.unpack('<21s',data[2:23])
    filename = dat[0].strip()
    
    #metadata
    reals = {}
    
    formatstring = '<4i4s4s4s4s20s3s11s1s8s' #run
    formatstring += '60s4s4s4s4s3i4s4s2i6s6s' #sample
    formatstring += '6s4s4s4s4s4s4s2i4s4s4s4s4s4s4s' #det
    formatstring += '4s4s4s4s4s4s' #resolution
    formatstring += 'L2i' #tslice
    formatstring += 'L4s4s4s2i' #temp
    formatstring += '2L4s4s4s4s4s' #magnet
    formatstring += '4s4s' #bmstp
    formatstring += '3i4s4s4s4s42s' #params
    formatstring += 'L4s4si' #voltage
    formatstring += '2L4s4s' #polarization
    formatstring += '4i4s4s4s4s4s' #analysis
    
    #print formatstring
    
    #print struct.calcsize(formatstring)
    
    (metadata['run.npre'],
     metadata['run.ctime'],
     metadata['run.rtime'],
     metadata['run.numruns'],
     reals['run.moncnt'],
     reals['run.savmon'],
     reals['run.detcnt'],
     reals['run.atten'],
     metadata['run.datetime'],
     metadata['run.type'],
     metadata['run.defdir'],
     metadata['run.mode'],
     metadata['run.reserve'],
     metadata['sample.labl'],
     reals['sample.trns'],
     reals['sample.thk'],
     reals['sample.position'],
     reals['sample.rotang'],
     metadata['sample.table'],
     metadata['sample.holder'],
     metadata['sample.blank'],
     reals['sample.temp'],
     reals['sample.field'],
     metadata['sample.tctrlr'],
     metadata['sample.magnet'],
     metadata['sample.tunits'],
     metadata['sample.funits'],
     metadata['det.typ'],
     reals['det.calx1'],
     reals['det.calx2'],
     reals['det.calx3'],
     reals['det.caly1'],
     reals['det.caly2'],
     reals['det.caly3'],
     metadata['det.num'],
     metadata['det.spacer'],
     reals['det.beamx'],
     reals['det.beamy'],
     reals['det.dis'],
     reals['det.ang'],
     reals['det.siz'],
     reals['det.bstop'],
     reals['det.blank'],
     reals['resolution.ap1'],
     reals['resolution.ap2'],
     reals['resolution.ap12dis'],
     reals['resolution.lmda'],
     reals['resolution.dlmda'],
     reals['resolution.nlenses'],
     metadata['tslice.slicing'],
     metadata['tslice.multfact'],
     metadata['tslice.ltslice'],
     metadata['temp.printemp'],
     reals['temp.hold'],
     reals['temp.err'],
     reals['temp.blank'],
     metadata['temp.extra'],
     metadata['temp.reserve'],
     metadata['magnet.printmag'],
     metadata['magnet.sensor'],
     reals['magnet.current'],
     reals['magnet.conv'],
     reals['magnet.fieldlast'],
     reals['magnet.blank'],
     reals['magnet.spacer'],
     reals['bmstp.xpos'],
     reals['bmstp.ypos'],
     metadata['params.blank1'],
     metadata['params.blank2'],
     metadata['params.blank3'],
     reals['params.trsncnt'],
     reals['params.extra1'],
     reals['params.extra2'],
     reals['params.extra3'],
     metadata['params.reserve'],
     metadata['voltage.printvolt'],
     reals['voltage.volts'],
     reals['voltage.blank'],
     metadata['voltage.spacer'],
     metadata['polarization.printpol'],
     metadata['polarization.flipper'],
     reals['polarization.horiz'],
     reals['polarization.vert'],
     metadata['analysis.rows1'],
     metadata['analysis.rows2'],
     metadata['analysis.cols1'],
     metadata['analysis.cols2'],
     reals['analysis.factor'],
     reals['analysis.qmin'],
     reals['analysis.qmax'],
     reals['analysis.imin'],
     reals['analysis.imax']) = struct.unpack(formatstring,data[23:514])

    
    #Process reals into metadata
    for k,v in reals.items():
        #print k,type(v)
        metadata[k] = R4toFloat(v)
    
    dataformatstring = '<16401h'
    rawdata = numpy.array(struct.unpack(dataformatstring,data[514:]))
    
    detdata = numpy.empty(16384)
    
    ii=0
    skip=0
    while(ii < 16384):
        if(((ii+skip) %1022)==0):
            skip+=1
        detdata[ii] = I2Decompress(rawdata[ii+skip])
        ii+=1
    
    detdata.resize(128,128)
    
    return (detdata,metadata)

def I2Decompress(val):
    """Take a 'compressed' I*2 value and convert to I*4.
       
       Code taken from IGOR Pro macros by SRK. VAX Fortran code is ultimate source (RW_DATAFILE.FOR)"""
    
    ib=10
    nd=4
    ipw = math.pow(ib,nd)
    if val <= -ipw:
        npw=trunc(-val/ipw)
        val=(-val % ipw)*(math.pow(ib,npw))
        return val
    else:
        return val

def arrayI2Decompress(datarray):
    """Apply the I2 to I4 decompression routine to a whole array"""
    for element in datarray.flat:
        element = I2Decompress(element)
    
    return datarray


def trunc(val):
    """Return the integer closest to the supplied value in the direction of zero"""
    
    if val < 0:
        return math.ceil(val)
    elif val > 0:
        return math.floor(val)
    else:
        return val

if __name__ == '__main__':
    sensitivity = readNCNRSensitivity("test.div")
    #print sensitivity
    detdata,metadata=readNCNRData("MAY06001.SA3_CM_D545")
    #plt.figure()
    #plt.imshow(detdata)
    #plt.show()
    



import numpy as np
from uncertainties.unumpy import uarray, nominal_values as uval, std_devs as udev

from reductus.dataflow.lib import err1d

from .deadtime_fit import deadtime_from_counts, estimate_incident


class DeadTimeData(object):
    def __init__(self, datasets, tau_NP, tau_P,
                 attenuators, rates, index):
        self.warnings = []

        self.datasets = datasets
        self.tau_NP = tau_NP
        self.tau_P = tau_P
        self.attenuators = attenuators
        self.incident_beam = rates
        self.index = index

    def plot(self):
        raise NotImplementedError("see deadtime_fit for plots we want")

    def get_metadata(self):
        metadata =  {
            "tau_P": self.tau_P,
            "tau_NP": self.tau_NP,
            "attenuators": self.attenuators,
        }
        return metadata
    
    def get_plottable(self):
        return {"params": self.get_metadata(), "type": "params"}


def fit_dead_time(datasets, source='detector', mode='auto'):
    time = [data.monitor.count_time for data in datasets]
    if source == 'monitor':
        counts = [data.monitor.counts for data in datasets]
    elif source == 'detector':
        counts = [data.detector.counts for data in datasets]
    else:
        raise ValueError("Source should be detector or monitor")
    data = datasets[-1]
    index = ~data.mask if data.mask is not None else slice(None, None)
    pairs = [(c[index], t[index]) for c, t in zip(counts, time)]
    res = deadtime_from_counts(pairs, mode=mode)
    tau_NP, tau_P, attenuators, rates = res
    attenuators = 1.0/uarray(*attenuators)
    attenuators = list(zip(uval(attenuators), udev(attenuators)))
    dead_time = DeadTimeData(datasets, tau_NP, tau_P, attenuators, rates, index)

    return dead_time


def apply_monitor_dead_time(data, tau_NP=0.0, tau_P=0.0):
    m = data.monitor.counts
    t = data.monitor.count_time
    dm = np.sqrt(data.monitor.counts_variance
                 if data.monitor.counts_variance is not None else data.monitor.counts)
    I, dI = estimate_incident((m/t, dm/t),
                              tau_NP=[tau_NP, 0], tau_P=[tau_P, 0])
    data.monitor.counts, data.monitor.counts_variance = (I*t), (dI*t)**2


def apply_detector_dead_time(data, tau_NP=0.0, tau_P=0.0):
    m = data.detector.counts
    t = data.monitor.count_time
    dm = np.sqrt(data.detector.counts_variance
                 if data.detector.counts_variance is not None else data.detector.counts)
    I, dI = estimate_incident((m/t, dm/t),
                              tau_NP=[tau_NP, 0], tau_P=[tau_P, 0])
    data.detector.counts, data.detector.counts_variance = (I*t), (dI*t)**2


def saturation_correction(counts, time, saturation):
    # type: (np.ndarray, np.ndarray, np.ndarray) -> Tuple[np.ndarray, np.ndarray, np.ndarray]
    rate = counts / time
    # TODO: assert that saturation is sorted by the first value
    if saturation is None:
        C, varC = counts, counts
        mask = (rate >= 0.)
    elif saturation.shape[0] == 3:
        E, varE = err1d.interp(rate, saturation[0], saturation[1],
                               saturation[2]**2)
        C, varC = err1d.div(counts, counts, E, varE)
        mask = (rate <= saturation[0, -1])
    else:
        E = np.interp(rate, saturation[0], saturation[1])
        C, varC = counts/E, counts/E**2
        mask = (rate <= saturation[0, -1])

    return C, varC, mask


def apply_monitor_saturation(data):
    m = data.monitor.counts
    t = data.monitor.count_time
    I, varI, mask = saturation_correction(m, t, data.monitor.saturation)
    data.monitor.counts, data.monitor.counts_variance = I, varI
    # If any points exceed the saturation calibration, mask them.
    if not mask.all():
        data.mask = mask & (True if data.mask is None else data.mask)


def apply_detector_saturation(data):
    t = data.monitor.count_time
    m = data.detector.counts
    I, varI, mask = saturation_correction(m, t, data.detector.saturation)
    data.detector.counts, data.detector.counts_variance = I, varI
    # If any points exceed the saturation calibration, mask them.
    if not mask.all():
        data.mask = mask & (True if data.mask is None else data.mask)




#!/usr/bin/env python
from copy import copy

import numpy as np

from .refldata import ReflData, PSDData
from .nexusref import load_nexus_entries, nexus_common
from .nexusref import data_as, str_data
from .nexusref import TRAJECTORY_INTENTS
from .resolution import FWHM2sigma

def load_metadata(filename, file_obj=None):
    """
    Load the summary info for all entries in a NeXus file.
    """
    return load_nexus_entries(filename, file_obj=file_obj,
                              meta_only=True, entry_loader=NG7PSD)

def load_entries(filename, file_obj=None, entries=None):
    #print("loading", filename, file_obj)
    return load_nexus_entries(filename, file_obj=file_obj, entries=entries,
                              meta_only=False, entry_loader=NG7PSD)

WAVELENGTH = 4.768
WAVELENGTH_DISPERSION = 0.025
S1_DISTANCE = -1722.25
S2_DISTANCE = -222.25
S3_DISTANCE = 229.0
DETECTOR_DISTANCE = 2006.
PIXEL_WIDTH = 100./256
PIXEL_OFFSET = np.linspace(0, 100, 257)[:-1] - 50 + 1/256

class NG7PSD(PSDData):
    """
    NG7PSD entry.

    See :class:`refldata.ReflData` for details.
    """
    format = "NeXus"
    probe = "neutrons"

    def __init__(self, entry, entryname, filename):
        super(NG7PSD, self).__init__()
        nexus_common(self, entry, entryname, filename)
        self.geometry = 'horizontal'

    def load(self, entry):
        #print(entry['instrument'].values())
        das = entry['DAS_logs']
        n = self.points
        raw_intent = str_data(das, 'trajectoryData/_scanType')
        if raw_intent in TRAJECTORY_INTENTS:
            self.intent = TRAJECTORY_INTENTS[raw_intent]

        # Polarizers
        self.polarization = ''

        # Monochromator
        self.monochromator.wavelength = data_as(entry, 'instrument/monochromator/wavelength', 'Ang', rep=n, NA=None)
        self.monochromator.wavelength_resolution = data_as(entry, 'instrument/monochromator/wavelength_error', 'Ang', rep=n, NA=None)
        if self.monochromator.wavelength is None:
            self.warn("Wavelength is missing for %r; using %.3f A"
                      % (self.name, WAVELENGTH))
            self.monochromator.wavelength = WAVELENGTH
        if self.monochromator.wavelength_resolution is None:
            self.warn("Wavelength resolution is missing for %r; using %.1f%% dL/L FWHM"
                      % (self.name, 100*WAVELENGTH_DISPERSION))
            self.monochromator.wavelength_resolution = \
                FWHM2sigma(WAVELENGTH_DISPERSION*self.monochromator.wavelength)

        # Slits
        # Read slit opening from the data file
        for k, slit in enumerate((self.slit1, self.slit2, self.slit3)):
            x = 'slitAperture%d/softPosition'%(k+1)
            x_target = 'slitAperture%d/desiredSoftPosition'%(k+1)
            slit.x = data_as(das, x, 'mm', rep=n)
            slit.x_target = data_as(das, x_target, 'mm', rep=n)
            #y = 'vertSlitAperture%d/softPosition'%(k+1)
            #y_target = 'vertSlitAperture%d/desiredSoftPosition'%(k+1)
            #slit.y = data_as(das, y, 'mm', rep=n)
            #slit.y_target = data_as(das, y_target, 'mm', rep=n)
        # There is no slit 4: use detector pixel width instead.
        self.slit4.x = data_as(entry, 'instrument/PSD/x_pixel_size', 'mm')
        if self.slit4.x is None:
            self.slit4.x = 0.390625  # 10 cm / 256 pixels
        self.slit4.x_target = self.slit4.x
        # Get slit distances as set by nexus config.js
        self.slit1.distance = data_as(entry, 'instrument/presample_slit1/distance', 'mm')
        self.slit2.distance = data_as(entry, 'instrument/presample_slit2/distance', 'mm')
        self.slit3.distance = data_as(entry, 'instrument/predetector_slit1/distance', 'mm')
        self.slit4.distance = data_as(entry, 'instrument/predetector_slit2/distance', 'mm')
        # Fill in missing distance data, if they aren't in the config file
        for slit, default in (
                (self.slit1, S1_DISTANCE),
                (self.slit2, S2_DISTANCE),
                (self.slit3, S3_DISTANCE),
                (self.slit4, DETECTOR_DISTANCE),
                ):
            if slit.distance is None:
                slit.distance = default
        # TODO: Correct for distance increase with q (5mm over 2m at q=0.2)
        # ... and the corresponding 1/4 % decrease in divergence per pixel
        # As the sample drops the slits pull away from the sample according
        # to the following:
        #    d = x * sqrt(tan(theta)^2 + 1)  # distance to detector center
        #      = nominal_distance * sqrt(tan(radians(detector_angle))**2 + 1)
        #    delta = d - nominal_distance
        # so S1, S2, S3 and detector distance all need to increase by this much.
        # Note: This calculation is used when computing the property Td_target,
        # the pixel angle for each point on the detector.

        # Detector
        self.detector.wavelength = self.monochromator.wavelength
        self.detector.wavelength_resolution = self.monochromator.wavelength_resolution
        self.detector.deadtime = data_as(entry, 'instrument/PSD/dead_time', 'us')
        self.detector.deadtime_error = data_as(entry, 'instrument/PSD/dead_time_error', 'us')
        self.detector.distance = data_as(entry, 'instrument/PSD/distance', 'mm')
        self.detector.rotation = data_as(entry, 'instrument/PSD/rotation', 'degree')
        if self.detector.distance is None:
            self.detector.distance = DETECTOR_DISTANCE
            self.warn("PSD distance is missing for %r; using %.1f mm"
                      % (self.name, self.detector.distance))

        # TODO: ng7r/nexus/config.js needs to record info about the PSD

        # Load data from linear detector.  Note that counts/liveROI may not
        # match if counts/roiAgainst is against a different detector.
        self.detector.counts = data_as(das, 'linearDetector/counts', '', dtype='d')
        #print("detector shape", self.detector.counts.shape)
        self.detector.counts_variance = self.detector.counts.copy()
        self.detector.dims = self.detector.counts.shape[1:]
        npixels = self.detector.dims[0]
        # TODO: is detector center in the data file?
        self.detector.center = [128, 0]
        self.detector.width_x = data_as(entry, 'instrument/PSD/x_pixel_size', 'mm', rep=npixels)
        self.detector.offset_x = data_as(entry, 'instrument/PSD/x_pixel_offset', 'mm', rep=npixels)
        self.detector.mask = data_as(entry, 'instrument/PSD/pixel_mask', '')
        if self.detector.width_x is None:
            self.detector.width_x = np.full(npixels, PIXEL_WIDTH)
        if self.detector.offset_x is None:
            self.detector.offset_x = PIXEL_OFFSET
        if self.detector.mask is None:
            self.detector.mask = np.array([1] + [0]*(npixels-1))

        # Angles
        if 'q' not in das:
            raise ValueError("Unknown sample angle in file %r" % self.name)
        self.Qz_target = data_as(das, 'q/z', '', rep=n)
        # Ignore sampleTilt for now since it is arbitrary.  NG7 is not
        # using zeros for the sampleTilt motor in a predictable way.
        #tilt = data_as(das, 'sampleTilt/softPosition', 'degree', rep=n)
        tilt = 0.
        # TODO: is thetaIncident tied to sample elevation actual or target?
        # We could recompute theta from sample elevation softPosition and
        # desiredSoftPosition respectively if we care.
        theta = data_as(das, 'q/thetaIncident', 'degree', rep=n)
        self.sample.angle_x = theta + tilt
        self.detector.angle_x = 2*theta
        #tilt_target = data_as(das, 'sampleTilt/desiredSoftPosition', 'degree', rep=n)
        self.sample.angle_x_target = self.sample.angle_x
        self.detector.angle_x_target = self.detector.angle_x

        # TODO: what does angular resolution mean before integration?
        from .angles import divergence_simple as divergence
        # Use slit2 and pixel width for the collimation geometry. The values
        # for pixel width and distances are stored as slit 4.
        slits = (self.slit4.x, self.slit2.x)
        distance = (self.slit4.distance, self.slit2.distance)
        theta = self.sample.angle_x
        self.angular_resolution = divergence(
            slits=slits, distance=distance, T=theta,
            sample_width=np.inf,
            use_sample=False,
        )

    @property
    def Ti_target(self):
        return self.sample.angle_x[:, None] # [n, 1]

    @property
    def Td_target(self):
        # Assuming the detector tilts with angle as the sampleElevator drops,
        # increasing the length of the arm slightly at the sample center.
        # Note: using a fixed distance would give 2% error on angle at q=0.5
        x = self.detector.distance  # [1]
        theta = np.radians(self.detector.angle_x) # [n]
        distance = x * np.sqrt(np.tan(theta)**2 + 1)  # [n] * [1] = [n]
        center = self.detector.center[0]  # [1]
        offset = self.detector.offset_x # [p]
        pixel_y = offset + center # [p]
        pixel_theta = np.arctan2(pixel_y[None, :], distance[:, None]) # [n, p]
        return np.degrees(pixel_theta)

    @property
    def Ti(self):
        return self.Ti_target

    @property
    def Td(self):
        return self.Td_target

    @property
    def Li(self):
        return self.monochromator.wavelength[:, None]

    @property
    def Ld(self):
        return self.detector.wavelength[:, None]

def apply_integration(
        data, spec=(1, 0), left=(1, 0), right=(1, 0), pixel_range=(1, 256),
        degree=0, mc_samples=0, seed=None, slices=[],
    ):
    """
    Apply integration to *data* returning specular, backgroundm, residual
    after background subtraction and slices.

    *data* is a dataset containing psd data in *data.v*.

    *spec*, *left* and *right* are (scale, offset) for the boundaries of the
    specular, left background and right background respectively.  The scale
    applies to the divergence estimated from slit 1 and the offset is the
    number of pixels to add or subtract from that width.

    *pixel_range* refers to the outer edges of the valid detector pixels.

    *degree* is the polynomial degree fitted to the background.

    *mc_samples* are the number of Monte Carlo samples to use when estimating
    background value and uncertainty under the signal.  Use 0 for an estimate
    directly from the uncertainty in the fitting parameters
    (**mc_samples==0 is broken** as of this writing).  Set *seed* to a fixed
    value for reproducible results.

    *slices* is a list of coordinates at which to display cross sections of
    the background fit.
    """
    from reductus.dataflow.lib.seed import push_seed

    # Make sure Monte Carlo simulations are reproducible by using the
    # same seed every time.
    with push_seed(seed):
        Is, Ib, residual, slice_plot = integrate(
            data, spec, left, right, pixel_range,
            degree, mc_samples, slices,
        )

    spec_data, back_data = _build_1d(data, Is), _build_1d(data, Ib)

    resid_data = copy(data)
    resid_data.v = residual

    return spec_data, back_data, resid_data, slice_plot

def _build_1d(head, intensity):
    # Initialize new from existing data
    data = ReflData()
    for field_name in data._fields:
        setattr(data, field_name, getattr(head, field_name, None))
    for group_name, _ in data._groups:
        data_group, head_group = getattr(data, group_name), getattr(head, group_name)
        for field_name in data_group._fields:
            setattr(data_group, field_name, getattr(head_group, field_name, None))

    # Kill fields we don't want anymore
    v, dv = intensity
    data.detector.counts = v
    data.detector.counts_variance = dv**2
    data.detector.dims = [1]
    data.detector.center = [0, 0]
    data.detector.width_x = np.sum(data.detector.width_x)
    data.detector.offset_x = 0.
    data.detector.mask = None
    data.warnings = []  # initialize per-file history
    data.intent = head.intent
    return data

def integrate(data, spec, left, right, pixel_range,
              degree, mc_samples, slices):
    from reductus.dataflow.lib import err1d
    from reductus.dataflow.lib.wsolve import wpolyfit

    nframes, npixels = data.v.shape

    # Determine the boundaries of each frame
    slit_width, pixel_width = data.slit1.x, data.slit4.x
    divergence = slit_width / pixel_width
    spec_width = spec[0]*divergence + spec[1]
    back_left = spec_width + np.maximum(left[0]*divergence + left[1], 0)
    back_right = spec_width + np.maximum(right[0]*divergence + right[1], 0)
    #print("integrate", spec, slit_width, pixel_width, spec_width, back_left, back_right)

    # Pixels are numbered from 1 to npixels, including the center pixel.
    # Normalize the pixels so the polynomial fitter uses smaller x.
    # Could also normalize by resolution width, but that makes it harder
    # to identify where the integration region is bad.
    center = data.detector.center[0]
    min_pixel, max_pixel = pixel_range
    p1 = np.maximum(min_pixel-center, -back_left)
    p2 = np.maximum(min_pixel-center, -spec_width)
    p3 = np.minimum(max_pixel-center, +spec_width)
    p4 = np.minimum(max_pixel-center, +back_right)
    pixel = np.arange(1, npixels+1) - center

    # TODO: maybe estimate divergence from specular peak width?
    # The following does not work:
    #    # Assume signal width is chosen to span +/- 2 sigma, or 95%.
    #    sigma = (p3 - p2)/4 * pixel_width
    #    divergence = np.degrees(np.arctan(sigma / data.detector.distance))

    # Find slices we want to plot by looking up the selected slice values
    # in the list of y values for the frames.
    (_, _), (yaxis, _) = data.get_axes() # get the data axes
    index = np.searchsorted(yaxis, slices)
    # TODO: search y-axis as bin edges rather than centers
    index = index[(index > 0)&(index < nframes)] - 1 # can't do last slice
    index = set(index) # duplicates and order don't matter for sets

    # Prepare plottable for slices
    series = [] # [{"label": str}, ...]
    lines = [] # [{c: {"values": [], "errorbars": []} for c in (pixel, intensity, error)}
    columns = {
        "pixel": {"label": "Pixel", "units": ""}, # no errorbars
        "intensity": {"label": "Intensity", "units": "counts", "errorbars": "error"}
    }
    plottable = {
        "type": "1d",
        "title": data.name + ":" + data.entry,
        "entry": data.entry,
        "columns": columns,
        "options": {
            "series": series,
            "axes": {
                "xaxis": {"label": "Pixel"},
                "yaxis": {"label": "Intensity (counts)"},
            },
            "xcol": "pixel",
            "ycol": "intensity",
            "errorbar_width": 0,
        },
        "data": lines,
    }
    def addline(label, x, y, dy=None):
        #print("line", label, x, y, dy)
        if dy is not None:
            line = [[xk, yk, {"yupper":yk+dyk, "ylower": yk-dyk}]
                    for xk, yk, dyk in zip(x, y, dy)]
        else:
            line = [[xk, yk] for xk, yk in zip(x, y)]
        series.append(label)
        lines.append(line)

    # Cycle through detector frames gathering signal and background for each.
    results = []
    for k in range(nframes):
        # Get data for frame.
        y, dy = data.v[k], data.dv[k]
        spec_idx = (pixel >= p2[k]) & (pixel <= p3[k])
        full_idx = (pixel >= p1[k]) & (pixel <= p4[k])
        back_idx = full_idx & ~spec_idx
        spec_x, spec_y, spec_dy = pixel[spec_idx], y[spec_idx], dy[spec_idx]
        back_x, back_y, back_dy = pixel[back_idx], y[back_idx], dy[back_idx]

        if not len(spec_x) or not len(back_x):
            results.append((np.nan, np.nan, np.nan, np.nan, np.zeros_like(y)))
            continue

        # Integrate frame data.
        # TODO: Could do sub-pixel interpolation at the boundary?
        Is, dIs = poisson_sum(spec_y, spec_dy)
        fit = wpolyfit(back_x, back_y, back_dy, degree=degree)
        # Uh, oh! Correlated errors on poly coefficients! How do we integrate?
        if mc_samples > 0: # using monte-carlo sampling
            # Generate a random set of polynomials from the fit
            coeffs = fit.rand(size=mc_samples)
            # Use Horner's method to evaluate all p_k over all x points
            px = np.zeros((len(spec_x), mc_samples))
            for c in coeffs.T:
                px *= spec_x[:, None]
                px += c[None, :]
            # Integrate p(x) over x
            integral = np.sum(px, axis=0)
            #print("integral", px.shape, mc_samples, coeffs.shape, integral.shape)
            # Find mean and variance of the integrated values
            Ib, dIb = np.mean(integral), np.std(integral)
        else: # using simple sum ignoring correlation in uncertainties
            est_y = fit(spec_x)
            est_dy = fit.ci(spec_x)
            Ib, dIb = err1d.sum(est_y, est_dy)

        # TODO: consider fitting gaussian to peak or finding FWHM of spec-back
        #signal = spec_y - fit(spec_x)
        #halfmax = signal.max() / 2
        #top_half = spec_x[signal > halfmax]
        #FWHM = top_half[-1] - top_half[0]
        #sigma = FWHM/(2*sqrt(2*log(2)))

        # add slices if the index is in the set of selected indices
        if k in index:
            valstr = str(yaxis[k])
            addline('data:'+valstr, pixel, y, dy)
            addline('spec:'+valstr, spec_x, fit(spec_x))
            addline('back:'+valstr, back_x, fit(back_x))

        # Show background residuals
        residual = y - fit(pixel)
        residual[~full_idx] = np.nan #0.
        #residual[spec_idx] += signal_jump

        results.append((Is, dIs, Ib, dIb, residual))

    Is, dIs, Ib, dIb, residual = (np.asarray(v) for v in zip(*results))

    return (Is, dIs), (Ib, dIb), residual, plottable

def poisson_sum(v, dv):
    """
    Sum data with poisson uncertainties.

    This happens to be the same as the sum of gaussian distributed
    uncertainties, with V = sum(v) and dV = sqrt(sum(dv^2)), preserving
    the relationship that variance of the sum matches its expected value.
    This even works if all elements of the sum have been scaled, for example
    by a shared deadtime factor.  In that case, Guassian propagation causes
    dv to be scaled by C, so variance is scaled by C^2, which becomes
    C dV after the square root.

    When integrating background, be sure the variance uses sqrt(counts)
    even when counts are zero for a bin, otherwise the uncertainty on the
    integrated counts will be overestimated.

    There is an important case where poisson integration will produce a
    different result from strict gaussian error propagation. That is for
    detectors where events give rise to a signal that is assigned to one
    pixel or its neighbour based on properties of the signal.  Inhomogeneities
    will lead to differing pixel widths across the detector, which shows
    up as different count rates per pixel in a flood-field measurement.
    In this case, it is better to sum the raw counts across the pixel region
    and scale by the average width of the contributing pixels rather than
    normalizing counts by pixel width then summing.

    For independent detectors with separate detection efficiency the
    situation is a little different.  In that case it is better to sum
    the efficiency corrected data using gaussian error propagation on
    the values, at least when the number of counts is large.  Not sure
    how to properly normalize across different detector efficiencies when
    counts are mostly zero, though gaussian approximation to the error
    is probably going to be good enough.

    When computing count rate, you should integrate peaks and backgrounds
    before normalizing by time or monitor counts. The normalization factor
    only drops out in the integration equations when you are using
    variance equal to number of counts.  However, when you are computing
    count rates, you will probably set the uncertainty for zero counts
    to 1/t rather than 0/t.  If you don't you will not be able to fit
    the data since the weighted residual will be infinite when you have
    zero uncertainty on a measurement.
    """
    return np.sum(v), np.sqrt(np.sum(dv**2))


if __name__ == "__main__":
    # Example:
    #   python -m reflred.ng7psd ncnr://ncnrdata/ng7/202001/27596/data/5ppm_NRW_0M33037.nxz.ng7
    from .nexusref import demo
    demo(loader=load_entries)



from reductus.dataflow import core as df
from reductus.dataflow.automod import make_modules, make_template, auto_module, get_modules
from reductus.dataflow.calc import process_template
from reductus.dataflow.data import Plottable
from reductus.dataflow.lib.exporters import exports_json

from . import steps
from . import templates
from reductus.reflred.refldata import ReflData
from reductus.reflred.footprint import FootprintData

INSTRUMENT = "xrr"

def define_instrument():
    # Define modules
    actions = get_modules(steps)
    modules = make_modules(actions, prefix=INSTRUMENT+'.')
    # Define data types
    refldata = df.DataType(INSTRUMENT+".refldata", ReflData)
    footprint = df.DataType("ncnr.refl.footprint.params", FootprintData)
    plottable = df.DataType(INSTRUMENT+".plot", Plottable)

    # Define instrument
    xrr = df.Instrument(
        id=INSTRUMENT,
        name='X-ray reflectometer',
        menu=[('steps', modules)],
        datatypes=[
            refldata, footprint, plottable,
            ],
        template_defs=df.load_templates(templates),
        )

    # Register instrument
    df.register_instrument(xrr)
    return xrr



"""
VSANS data loader
================

Load VSANS NeXus file into :mod:`vsansred.vsansdata` data structure.
"""
import io
from zipfile import ZipFile, is_zipfile
from collections import OrderedDict
import datetime
import h5py
import numpy as np

from reductus.dataflow.lib import hzf_readonly_stripped as hzf
from reductus.dataflow.lib import unit
from reductus.dataflow.lib.h5_open import h5_open_zip

from .vsansdata import VSansData, RawVSANSData, _s, _b

metadata_lookup = OrderedDict([
    #"det.dis", "DAS_logs/detectorPosition/softPosition",
    
    #"det.beamx", "instrument/detector/beam_center_x",
    #"det.beamy", "instrument/detector/beam_center_y",
    #"det.pixeloffsetx", "instrument/detector/x_offset",
    #"det.pixelsizex", "instrument/detector/x_pixel_size",
    #"det.pixeloffsety", "instrument/detector/y_offset",
    #"det.pixelsizey", "instrument/detector/y_pixel_size",
    ("run.filename", "DAS_logs/trajectoryData/fileName"),
    ("analysis.intent", "DAS_logs/trajectoryData/intent"),
    ("analysis.filepurpose", "DAS_logs/trajectoryData/filePurpose"),
    ("sample.name", "DAS_logs/sample/name"),
    #("sample.description", "DAS_logs/sample/description"),
    ("sample.labl", "DAS_logs/sample/description"), # compatibility
    ("resolution.lmda" , "instrument/beam/monochromator/wavelength"),
    ("resolution.dlmda", "instrument/beam/monochromator/wavelength_spread"),
    ("m_det.beamx", "DAS_logs/middleRightAreaDetector/beamCenterX"),
    ("m_det.beamy", "DAS_logs/middleRightAreaDetector/beamCenterY"),
    ("m_det.dis", "DAS_logs/geometry/sampleToMiddleRightDetector"),
    ("f_det.beamx", "DAS_logs/frontRightAreaDetector/beamCenterX"),
    ("f_det.beamy", "DAS_logs/frontRightAreaDetector/beamCenterY"),
    ("f_det.dis", "DAS_logs/geometry/sampleToFrontRightDetector"),
    ("m_det_des.dis", "DAS_logs/carriage2Trans/desiredSoftPosition"),
    ("f_det_des.dis", "DAS_logs/carriage1Trans/desiredSoftPosition"),
    ("polarization.front", "DAS_logs/frontPolarization/direction"),
    ("polarization.back", "DAS_logs/backPolarization/direction"),
    ("polarization.backname", "DAS_logs/backPolarization/name"),
    ("polarization.backstart", "DAS_logs/backPolarization/timestamp"),
    ("run.filePrefix", "DAS_logs/trajectoryData/filePrefix"),
    ("run.instFileNum", "DAS_logs/trajectoryData/instFileNum"),
    ("run.pointnum", "DAS_logs/trajectoryData/pointNum"),
    #("run.detcnt", "control/detector_counts"),
    ("run.rtime", "control/count_time"),
    ("run.moncnt", "control/monitor_counts"),
    ("run.atten", "instrument/attenuator/num_atten_dropped"),
    ("analysis.groupid", "DAS_logs/trajectoryData/groupid"),
    ("run.configuration", "DAS_logs/configuration/key"),
    ("sample.thk", "DAS_logs/sample/thickness"),
    ("adam.voltage", "DAS_logs/adam4021/voltage"),
    ("sample.temp", "DAS_logs/temp/primaryNode/average_value"),
    ("sample_des.temp", "DAS_logs/temp/desiredPrimaryNode"),
    ("resolution.ap1", "DAS_logs/geometry/sourceAperture"),
    ("resolution.ap2", "instrument/sample_aperture/size"),
    ("resolution.ap12dis", "instrument/source_aperture/distance"),
    ("resolution.guide", "DAS_logs/guide/guide"),
    ("sample.position", "instrument/sample_aperture/distance"),
    ("electromagnet_lrm.field","DAS_logs/electromagnet_lrm/field"),
    ("mag.value","DAS_logs/mag/value"),
    ("acamplitude.voltage",  "DAS_logs/acAmplitude/voltage"),
    ("waveformgenerator.frequency",  "DAS_logs/waveformGenerator/frequency"),
    ("rfflipperpowersupply.voltage",  "DAS_logs/RFFlipperPowerSupply/actualVoltage/average_value"),
    ("rfflipperpowersupply.frequency",  "DAS_logs/RFFlipperPowerSupply/frequency"),
    ("huberRotation.softPosition",  "DAS_logs/huberRotation/softPosition"),
    ("start_time","start_time"),
    ("end_time","end_time"),
    ("eventfile", "DAS_logs/areaDetector/eventFileName"),
    ("he3_back.opacity", "DAS_logs/backPolarization/opacityAt1Ang"),
    ("he3_back.te", "DAS_logs/backPolarization/glassTransmission"),
])

he3_metadata_lookup = OrderedDict([
    ("run.filename", "DAS_logs/trajectoryData/fileName"),
    ("sample.labl", "DAS_logs/sample/description"), # compatibility
    ("analysis.intent", "DAS_logs/trajectoryData/intent"),
    ("analysis.filepurpose", "DAS_logs/trajectoryData/filePurpose"),
    ("he3_back.starttime", "DAS_logs/backPolarization/timestamp"),
    ("he3_back.name", "DAS_logs/backPolarization/name"),
    ("he3_back.inbeam", "DAS_logs/backPolarization/inBeam"),
    ("he3_back.opacity", "DAS_logs/backPolarization/opacityAt1Ang"),
    ("he3_back.te", "DAS_logs/backPolarization/glassTransmission"),
    ("he3_back.direction", "DAS_logs/backPolarization/direction"),
    ("run.instrumentScanID", "DAS_logs/trajectory/instrumentScanID"),
    ("run.instFileNum", "DAS_logs/trajectoryData/instFileNum"),
    ("run.rtime", "control/count_time"),
    ("run.moncnt", "control/monitor_counts"),
    ("run.atten", "instrument/attenuator/num_atten_dropped"),
    ("sample.name", "DAS_logs/sample/name"),
    ("resolution.lmda" , "instrument/beam/monochromator/wavelength"),
    ("resolution.dlmda", "instrument/beam/monochromator/wavelength_spread"),
    ("m_det.dis_des", "DAS_logs/carriage2Trans/desiredSoftPosition"),
    ("f_det.dis_des", "DAS_logs/carriage1Trans/desiredSoftPosition"),
    ("start_time","start_time"),
    ("end_time","end_time"),
    ("eventfile", "DAS_logs/areaDetector/eventFileName")
])

unit_specifiers = {
    "det.dis": "cm",
    "det.pixelsizex": "cm",
    "det.pixeloffsetx": "cm",
    "det.pixelsizey": "cm",
    "det.pixeloffsety": "cm",
    "sample.thk": "cm",
    "resolution.ap1": "cm",
    "resolution.ap2": "cm"
}

def process_sourceAperture(field, units):
    import numpy as np
    def handler(v):
        if _s(v) == 'OUT':
            return v
        else:
            return float(v.split()[0])
    handle_values = np.vectorize(handler)
    value = handle_values(field[()])
    units_from = ""
    v0 = field[0].split()
    if _s(value[0]) == 'OUT':
        return value
    if len(v0) > 1:
        units_from = v0[1]
    if type(units_from) == bytes:
        units_from = units_from.decode('utf-8')
    converter = unit.Converter(units_from)
    return converter(value, units)    

def data_as(field, units):
    """
    Return value of field in the desired units.
    """
    if field.name.split('/')[-1] == 'sourceAperture':
        return process_sourceAperture(field, units)
    else:
        units_in = field.attrs.get('units', '')
        if type(units_in) == bytes:
            units_in = units_in.decode('utf-8')
        converter = unit.Converter(units_in)
        value = converter(field[()], units)
        return value

def load_detector(dobj, load_data=True):
    # load detector information from a NeXuS group
    detector = OrderedDict()
    for k in dobj:
        if not load_data and k == 'data':
            continue
        subobj = dobj[k]
        detector[k] = OrderedDict(value=subobj[()], attrs=OrderedDict(_toDictItem(subobj.attrs)))
        if hasattr(subobj, 'shape'):
            detector[k]['attrs']['shape'] = subobj.shape
        if hasattr(subobj, 'dtype'):
            detector[k]['attrs']['dtype'] = subobj.dtype.name
    return detector

def load_metadata(entry, multiplicity=1, i=1, metadata_lookup=metadata_lookup, unit_specifiers=unit_specifiers):
    metadata = OrderedDict()
    for mkey in metadata_lookup:
        field = entry.get(metadata_lookup[mkey], None)
        if field is not None:
            if mkey in unit_specifiers:
                field = data_as(field, unit_specifiers[mkey])
            else:
                field = field[()]
            if field.dtype.kind == 'f':
                field = field.astype("float")
            elif field.dtype.kind == 'i':
                field = field.astype("int")
        
            if len(field) == multiplicity:
                metadata[mkey] = field[i]
            else:
                metadata[mkey] = field
        else:
            metadata[mkey] = field
    return metadata

def readVSANSNexuz(input_file, file_obj=None, metadata_lookup=metadata_lookup, load_data=True):
    """
    Load all entries from the NeXus file into sans data sets.
    """
    datasets = []
    file = h5_open_zip(input_file, file_obj)
    for entryname, entry in file.items():
        #areaDetector = entry['data/areaDetector'].value
        #shape = areaDetector.shape
        #if len(shape) < 2 or len(shape) > 3:
        #    raise ValueError("areaDetector data must have dimension 2 or 3")
        #    return
        #if len(shape) == 2:
            # add another dimension at the front
        #    shape = (1,) + shape
        #    areaDetector = areaDetector.reshape(shape)
        
        multiplicity = 1
        for i in range(multiplicity):
            metadata = load_metadata(entry, multiplicity, i, metadata_lookup=metadata_lookup, unit_specifiers=unit_specifiers)
            #print(metadata)
            detector_keys = [n for n in entry['instrument'] if n.startswith('detector_')]
            detectors = dict([(k, load_detector(entry['instrument'][k], load_data=load_data)) for k in detector_keys])
            metadata['entry'] = entryname
            if metadata.get('sample.labl', None) is not None and metadata.get('run.configuration', None) is not None:
                metadata['sample.description'] = _s(metadata["sample.labl"]).replace(_s(metadata["run.configuration"]), "")
            if metadata.get('run.filename', None) is None:
                metadata['run.filename'] = input_file
            dataset = RawVSANSData(metadata=metadata, detectors=detectors)
            datasets.append(dataset)            

    return datasets

def _toDictItem(obj, convert_bytes=False):
    if isinstance(obj, np.integer):
        obj = int(obj)
    elif isinstance(obj, np.floating):
        obj = float(obj)
    elif isinstance(obj, np.ndarray):
        obj = obj.tolist()
    elif isinstance(obj, datetime.datetime):
        obj = [obj.year, obj.month, obj.day, obj.hour, obj.minute, obj.second]
    elif isinstance(obj, list):
        obj = [_toDictItem(a, convert_bytes=convert_bytes) for a in obj]
    elif isinstance(obj, dict):
        obj = OrderedDict([(k, _toDictItem(v, convert_bytes=convert_bytes)) for k, v in obj.items()])
    elif isinstance(obj, bytes) and convert_bytes == True:
        obj = obj.decode()
    return obj
